# class L2Regularization:
#     r"""
#     Implements :math:`L_2` (Tikhonov) regularization.
#     The penalization term in this type of regularization is the square of the
#     :math:`L_2` (Euclidean) norm of a linear operator applied to the function
#     or vector
#     .. math::
#             \lambda \| \Gamma x \|_2^2
#     where :math:`\Gamma` is the so called Tikhonov operator
#     (matrix for finite vectors) and :math:`\lambda` is a positive real number.
#     This linear operator can be an arbitrary Python callable that correspond
#     to a linear transformation. However, the
#     :doc:`operators </modules/misc/operators>` module
#     provides several common linear operators.
#     Parameters:
#         linear_operator: Linear operator used for regularization. By default
#             the second derivative, which is related with the function
#             curvature, is penalized.
#         regularization_parameter: Scaling parameter (:math:`\lambda`) of the
#             penalization.
#     Examples:
#         Construct a regularization that penalizes the second derivative,
#         which is a measure of the curvature of the function.
#         >>> from skfda.misc.regularization import L2Regularization
#         >>> from skfda.misc.operators import LinearDifferentialOperator
#         >>>
#         >>> regularization = L2Regularization(
#         ...                     LinearDifferentialOperator(2),
#         ... )
#         By default the regularization penalizes the identity operator:
#         >>> regularization = L2Regularization()
#         Construct a regularization that penalizes the difference between
#         the points :math:`f(1)` and :math:`f(0)` of a function :math:`f`.
#         >>> regularization = L2Regularization(lambda x: x(1) - x(0))
#         Construct a regularization that penalizes the harmonic acceleration
#         operator :math:`Lf = \omega^2 D f + D^3 f`, that, when the
#         regularization parameter is large, forces the function to be
#         :math:`f(t) = c_1 + c_2 \sin \omega t + c_3 \cos \omega t`, where
#         :math:`\omega` is the angular frequency. This is useful for some
#         periodic functions.
#         >>> import numpy as np
#         >>>
#         >>> period = 1
#         >>> w = 2 * np.pi / period
#         >>> regularization = L2Regularization(
#         ...                     LinearDifferentialOperator([0, w**2, 0, 1]),
#         ... )
#     """
#
#     def __init__(
#         self,
#         linear_operator: Optional[Operator[OperatorInput, Any]] = None,
#         *,
#         regularization_parameter: float = 1,
#     ) -> None:
#         self.linear_operator = linear_operator
#         self.regularization_parameter = regularization_parameter
#
#     def penalty_matrix(
#         self,
#         basis: OperatorInput,
#     ) -> NDArrayFloat:
#         """Return a penalty matrix for ordinary least squares."""
#         linear_operator = (
#             Identity()
#             if self.linear_operator is None
#             else self.linear_operator
#         )
#
#         return self.regularization_parameter * gram_matrix(
#             linear_operator,
#             basis,
#         )
#
#
# class TikhonovRegularization(
#     L2Regularization[OperatorInput],
# ):
#
#     def __init__(
#         self,
#         linear_operator: Optional[Operator[OperatorInput, Any]] = None,
#         *,
#         regularization_parameter: float = 1,
#     ) -> None:
#
#         warnings.warn(
#             "Class TikhonovRegularization is deprecated. Use class "
#             "L2Regularization instead.",
#             DeprecationWarning,
#         )
#
#         return super().__init__(
#             linear_operator=linear_operator,
#             regularization_parameter=regularization_parameter,
#         )
#
#
# BasisTypes = Union[np.ndarray, FData, Basis]
# Regularization = L2Regularization[Any]
# RegularizationLike = Union[
#     None,
#     Regularization,
#     Iterable[Optional[Regularization]],
# ]
#
#
# def compute_penalty_matrix(
#     basis_iterable: Iterable[BasisTypes],
#     regularization_parameter: Union[float, Iterable[float]],
#     regularization: RegularizationLike,
# ) -> Optional[NDArrayFloat]:
#     """
#     Compute the regularization matrix for a linear differential operator.
#     X can be a list of mixed data.
#     """
#     # If there is no regularization, return 0 and rely on broadcasting
#     if regularization_parameter == 0 or regularization is None:
#         return None
#
#     # Compute penalty matrix if not provided
#     if not isinstance(regularization, Iterable):
#         regularization = (regularization,)
#
#     if not isinstance(regularization_parameter, Iterable):
#         regularization_parameter = itertools.repeat(
#             regularization_parameter,
#         )
#
#     penalty_blocks = [
#         np.zeros((len(b), len(b))) if r is None else
#         a * r.penalty_matrix(b)
#         for b, r, a in zip(
#             basis_iterable,
#             regularization,
#             regularization_parameter,
#         )]
#
#     return scipy.linalg.block_diag(  # type: ignore[no-any-return]
#         *penalty_blocks,
#     )


import numpy as np
from sklearn.preprocessing import MinMaxScaler


def basis_approximation_metric(derivation_coef: np.array,
                               metric_values: np.array,
                               regularization_coef: float = 0.7):

    # Get a sum of absolute values (L1 norm) for each polynom derivative
    sum_of_derivative_coef = np.array([sum(abs(x)) for x in derivation_coef])

    # Normalize metrics and der_coef
    normalized_derivative_coef = MinMaxScaler(feature_range=(0, 1)).fit_transform(
        sum_of_derivative_coef.reshape(-1, 1))[:, 0]
    normalized_metric_values = MinMaxScaler(feature_range=(0, 1)).fit_transform(
        metric_values.reshape(-1, 1))[:, 0]

    # Alternative approach where metric = regulizer*np.exp(der_coef)*metric_val
    # exponent_of_coef = np.array([np.exp(x) for x in normalized_derivative_coef])
    # polynom_metric = exponent_of_coef * self.mse

    # Basic approach where metric = regulizer*der_coef+metric_val
    polynom_metric = regularization_coef * normalized_derivative_coef + normalized_metric_values
    best_aprox_polynom_index = np.where(polynom_metric == min(polynom_metric))[0][0]

    return best_aprox_polynom_index



from typing import List, Mapping

from fedot_ind.core.models.detection.abstract_objects.AnomalyZone import AnomalyZone


class FileObject:
    """
    This class contains all data that contains in one file - time series, lables,
    transformation params, anomalies list
    """
    def __init__(self, 
            test_data: List[float], 
            filename: str
        ) -> None:
        # test data
        self.test_data = test_data
        # all data from file in form of dictioary
        self.time_series_data: Mapping[str, List[float]] = {}
        # list of nomalies in this data
        self.anomalies_list: List[AnomalyZone] = []
        # length of data in file
        self.data_length: int = None
        # filename and full path to file
        self.filename: str = filename
        #self.filepath: str = filepath
        # lables list(or dict???)
        self.lables: list = None
        self.elected_columns: List[str] = None
        self.current_elected_columns: List[str] = []

        # main transformed data
        self.transformed_data: Mapping[str, List[float]] = {}

        # additional ts that creates durnig analysis
        self.additional_min_ts: List[float] = None
        self.additional_max_ts: List[float] = None
        self.additional_distance_ts: List[float] = None
        self.additional_average_absolute_deviation_max: List[float] = None
        self.additional_average_absolute_deviation_sum: List[float] = None
        self.additional_transformed_average_absolute_deviation: List[float] = None
        self.additional_mean_ts: List[float] = None
        # threshold for <additional_transformed_average_absolute_deviation>
        self.threshold: float = None
        self.test_vector_ts = []

    def get_len_of_dataset(self) -> int:
        """
        This method need to make things more standart. 
        Returns len of data in this FileObject

        Returns:
            int: len of data
        """
        return len(self.test_data)
    
    def get_lables_list(self, min_heavines: int = 0) -> list:
        if len(self.lables):
            temp_list = []
            for predict in self.lables:
                if predict[2] > min_heavines:
                    temp_list.append(predict)
            return temp_list
        else:
            return False
    
    def get_lables_for_metrics(self, min_heavines: int = 0) -> List[int]:
        """
        Returns lables in suitable format for metrics: [0, 0, 0, 1, 1, 1, ..., 0]

        Args:
            min_heavines (int, optional): Set this arg to none-zero values if you 
            need to filter lables by heaviness of anomalies. Defaults to 0.

        Returns:
            List[int]: list in format [0, 0, 0, 1, 1, 1, ..., 0] with len the same 
            as len of data in this object
        """
        output_list = [0] * self.get_len_of_dataset()
        if len(self.lables):
            for predict in self.lables:
                if predict[2] > min_heavines:
                    begin = predict[0]
                    end = predict[1]
                    if begin < 0: begin = 0
                    if end >= self.get_len_of_dataset(): end = self.get_len_of_dataset() - 1 
                    for i in range(begin, end):
                        output_list[i] = 1
            return output_list
        else:
            return output_list
    
    def get_all_predicts_list_by_type(self, type_numbers: List[int]) -> List[AnomalyZone]:
        """
        This method returns list of predicts by their class
        or all predicts if type_numbers set to -1
        if type_number == -1 - returns all predicts
        Args:
            type_number (List[int]): could contain any of this numbers [-1, 0, 1, 2, 3]

        Returns:
            List[AnomalyZone]: list of anomalies
        """
        if not len(self.anomalies_list):
            return False
        else: 
            if -1 in type_numbers:
                predictions_list = []
                for predict in self.anomalies_list:
                    predictions_list.append(predict)
            else:
                predictions_list = []
                for predict in self.anomalies_list:
                    if predict.predicted_type in type_numbers: # and predict.heaviness >9: 
                    #if predict.cluster_type == type_number: dataset_type predicted_type
                        predictions_list.append(predict)
            return predictions_list


    def get_predict_for_metrics(self, classes: List[int] =[2, 3]) -> List[int]:
        """
        Returns predicts in suitable format for metrics: [0, 0, 0, 1, 1, 1, ..., 0]
        arg <classes> sets classes of predicts that will be included into output

        Args:
            classes (List[int], optional): could be any of [0, 1, 2, 3]. Defaults to [2, 3].

        Returns:
            List[int]: list in format [0, 0, 0, 1, 1, 1, ..., 0] with len the same 
            as len of data in this object
        """
        time_series_of_predict_for_metrics = [0] * self.get_len_of_dataset()

        if len(self.anomalies_list) == 0:
            return time_series_of_predict_for_metrics
        else:
            for zone in self.anomalies_list:
                if zone.predicted_type in classes:
                    for j in range(zone.get_start(), zone.get_end(), 1):
                        if 0 <= j < self.get_len_of_dataset():
                            time_series_of_predict_for_metrics[j] = 1
        return time_series_of_predict_for_metrics

    def copy_parts_of_datasets_in_anomaly_zones(self) -> None:
        """
        This method copy parts of raw data to anomaly zones
        """
        for i in range(len(self.anomalies_list)):
            if len(self.time_series_data.keys()):
                for column in list(self.time_series_data.keys()):
                    # Copy data of each time series
                    self.anomalies_list[i].data[column] = \
                        self.time_series_data[column][
                            self.anomalies_list[i].get_start():self.anomalies_list[i].get_end()
                            ]
            # Copy additional_min_ts
            if self.additional_min_ts  is not None:
                self.anomalies_list[i].min_data = self.additional_min_ts[
                    self.anomalies_list[i].get_start():self.anomalies_list[i].get_end()
                ]
            # Copy additional_max_ts
            if self.additional_max_ts is not None:
                self.anomalies_list[i].max_data = self.additional_max_ts[
                    self.anomalies_list[i].get_start():self.anomalies_list[i].get_end()
                ]
            # Copy distance_data
            if self.additional_distance_ts is not None:
                self.anomalies_list[i].distance_data = \
                    self.additional_distance_ts[
                            self.anomalies_list[i].get_start():self.anomalies_list[i].get_end()
                        ]
            if self.additional_average_absolute_deviation_max is not None:
            # Copy average_absolute_deviation
                self.anomalies_list[i].average_absolute_deviation = \
                    self.additional_average_absolute_deviation_max[
                            self.anomalies_list[i].get_start():self.anomalies_list[i].get_end()
                        ]
            # Copy average_absolute_deviation_transformed
            if self.additional_transformed_average_absolute_deviation is not None:
                self.anomalies_list[i].average_absolute_deviation_transformed = \
                    self.additional_transformed_average_absolute_deviation[
                            self.anomalies_list[i].get_start():self.anomalies_list[i].get_end()
                        ]

            # Copy additional_mean_ts
            if self.additional_mean_ts is not None:
                self.anomalies_list[i].additional_mean_ts = \
                    self.additional_mean_ts[
                            self.anomalies_list[i].get_start():self.anomalies_list[i].get_end()
                        ]
            # Copy threshold
            self.anomalies_list[i].threshold = self.threshold



from datetime import datetime
"""
    This is getter for formated current time for logging
"""

def get_current_time() -> str:
    now = datetime.now()
    return now.strftime("%H:%M:%S")


def time_now() -> str:
    now = datetime.now()
    return now.strftime("%H:%M:%S")



import logging
from typing import List, Union

import numpy as np
import pandas as pd
from fedot.api.main import Fedot
from fedot.core.pipelines.pipeline import Pipeline

from fedot_ind.api.utils.reader_collections import DataReader, YamlReader
from fedot_ind.api.utils.reporter import ReporterTSC
from fedot_ind.core.architecture.settings.task_factory import TaskGenerator
from fedot_ind.core.architecture.utils.utils import default_path_to_save_results


class FedotIndustrial(Fedot):
    """This class is used to run Fedot in industrial mode as FedotIndustrial.

    Args:
        input_config: dictionary with the parameters of the experiment.
        output_folder: path to the folder where the results will be saved.

    Example:
        First, configure experiment and instantiate FedotIndustrial class::

            industrial = FedotIndustrial(task='ts_classification',
                                         dataset='ItalyPowerDemand',
                                         strategy='topological',
                                         use_cache=False,
                                         timeout=15,
                                         n_jobs=2,
                                         logging_level=20)

        Next, download data::

            train_data, test_data, _ = industrial.reader.read(dataset_name='ItalyPowerDemand')

        Finally, fit the model and get predictions::

            model = industrial.fit(train_features=train_data[0], train_target=train_data[1])
            labels = industrial.predict(test_features=test_data[0])
            probs = industrial.predict_proba(test_features=test_data[0])
            metric = industrial.get_metrics(target=test_data[1], metric_names=['f1', 'roc_auc'])

    """

    def __init__(self, **kwargs):
        super(Fedot, self).__init__()

        self.logger = logging.getLogger('FedotIndustrialAPI')

        self.reporter = ReporterTSC()
        self.YAML = YamlReader()
        self.reader = DataReader()

        self.config_dict = None
        self.output_folder = kwargs.get('output_folder', None)

        self.__init_experiment_setup(**kwargs)
        self.solver = self.__init_solver()

    def __init_experiment_setup(self, **kwargs):
        self.logger.info('Initialising experiment setup')
        if not self.output_folder:
            self.output_folder = default_path_to_save_results()
        self.reporter.path_to_save = self.output_folder

        self.config_dict = self.YAML.init_experiment_setup(**kwargs)

    def __init_solver(self):
        self.logger.info('Initialising solver')

        if self.config_dict['task'] == 'ts_classification':
            if self.config_dict['strategy'] == 'fedot_preset':
                solver = TaskGenerator[self.config_dict['task']].value['fedot_preset']
            elif self.config_dict['strategy'] is None:
                solver = TaskGenerator[self.config_dict['task']].value['nn']
            else:
                solver = TaskGenerator[self.config_dict['task']].value['default']

        else:
            solver = TaskGenerator[self.config_dict['task']].value[0]

        return solver(self.config_dict)

    def fit(self,
            train_features: pd.DataFrame,
            train_target: np.ndarray,
            **kwargs) -> Pipeline:
        """
        Method for training Industrial model.

        Args:
            train_features: raw time series data
            train_target: target labels
            kwargs: additional parameters for solver, for example ``baseline_type`` – a model that could be selected
                    instead of Fedot pipeline

        Returns:
            :class:`Pipeline` object.

        """

        fitted_pipeline = self.solver.fit(train_ts_frame=train_features,
                                          train_target=train_target,
                                          **kwargs)
        return fitted_pipeline

    def predict(self,
                test_features: pd.DataFrame,
                **kwargs) -> np.ndarray:
        """
        Method to obtain prediction labels from trained Industrial model.

        Args:
            test_features: raw test data

        Returns:
            the array with prediction values

        """
        return self.solver.predict(test_features=test_features, **kwargs)

    def predict_proba(self,
                      test_features: pd.DataFrame,
                      **kwargs) -> np.ndarray:
        """
        Method to obtain prediction probabilities from trained Industrial model.

        Args:
            test_features: raw test data

        Returns:
            the array with prediction probabilities

        """
        return self.solver.predict_proba(test_features=test_features, **kwargs)

    def get_metrics(self,
                    target: Union[np.ndarray, pd.Series] = None,
                    metric_names: Union[str, List[str]] = ('f1', 'roc_auc', 'accuracy', 'logloss', 'precision'),
                    **kwargs) -> dict:
        """
        Method to obtain Gets quality metrics

        Args:
            target: target values
            metric_names: list of metric names desired to be calculated
            **kwargs: additional parameters

        Returns:
            the dictionary with calculated metrics

        """
        return self.solver.get_metrics(target, metric_names)

    def save_predict(self, predicted_data: Union[pd.DataFrame, np.ndarray], **kwargs) -> None:
        """
        Method to save prediction locally in csv format

        Args:
            predicted_data: predicted data. For TSC task it could be either labels or probabilities

        Returns:
            None

        """
        kind = kwargs.get('kind')
        self.solver.save_prediction(predicted_data, kind=kind)

    def save_metrics(self, metrics: dict) -> None:
        """
        Method to save metrics locally in csv format

        Args:
            metrics: dictionary with calculated metrics

        Returns:
            None
        """
        self.solver.save_metrics(metrics)

    def load(self, path):
        """Loads saved Industrial model from disk

        Args:
            path (str): path to the model
        """
        # self.pipeline.load(path)
        raise NotImplementedError()

    def plot_prediction(self, **kwargs):
        """Plot prediction of the model"""
        raise NotImplementedError()

    def explain(self, **kwargs):
        raise NotImplementedError()



from typing import Union

import numpy as np
import pandas as pd
from scipy.linalg import hankel


class HankelMatrix:
    """
    This class implements an algorithm for converting an original time series into a Hankel matrix.
    """

    def __init__(self,
                 time_series: Union[pd.DataFrame, pd.Series, np.ndarray, list],
                 window_size: int = None):
        self.__time_series = time_series
        self.__convert_ts_to_array()
        self.__window_length = window_size

        if len(self.__time_series.shape) > 1:
            self.__ts_length = self.__time_series[0].size
        else:
            self.__ts_length = self.__time_series.size

        if self.__window_length is None:
            self.__window_length = round(self.__ts_length * 0.35)
        else:
            self.__window_length = round(self.__ts_length * self.__window_length // 100)
        self.__subseq_length = self.__ts_length - self.__window_length + 1

        self.__check_windows_length()
        self.__trajectory_matrix = self.__get_trajectory_matrix()

    def __check_windows_length(self):
        if not 2 <= self.__window_length <= self.__ts_length / 2:
            raise ValueError("The window length must be in the interval [2, N/2].")

    def __convert_ts_to_array(self):
        if type(self.__time_series) == pd.DataFrame:
            self.__time_series = self.__time_series.values.reshape(-1, 1)
        elif type(self.__time_series) == list:
            self.__time_series = np.array(self.__time_series)
        else:
            self.__time_series = self.__time_series


    def __get_trajectory_matrix(self):
        if len(self.__time_series.shape) > 1:
            return [hankel(time_series[:self.__window_length + 1], time_series[self.__window_length:]) for time_series
                    in self.__time_series]
        else:
            return hankel(self.__time_series[:self.__window_length + 1], self.__time_series[self.__window_length:])

    @property
    def window_length(self):
        return self.__window_length

    @property
    def time_series(self):
        return self.__time_series

    @property
    def sub_seq_length(self):
        return self.__subseq_length

    @window_length.setter
    def window_length(self, window_length):
        self.__window_length = window_length

    @property
    def trajectory_matrix(self):
        return self.__trajectory_matrix

    @property
    def ts_length(self):
        return self.__ts_length

    @trajectory_matrix.setter
    def trajectory_matrix(self, trajectory_matrix: np.ndarray):
        self.__trajectory_matrix = trajectory_matrix


def get_x_y_pairs(train, train_periods, prediction_periods):
    """
    train_scaled - training sequence
    train_periods - How many data points to use as inputs
    prediction_periods - How many periods to ouput as predictions
    """
    train_scaled = train.T
    r = train_scaled.shape[0] - train_periods - prediction_periods
    x_train = [train_scaled[i:i + train_periods] for i in range(r)]
    y_train = [train_scaled[i + train_periods:i + train_periods + prediction_periods] for i in range(r)]

    # -- use the stack function to convert the list of 1D tensors
    # into a 2D tensor where each element of the list is now a row
    x_train = np.concatenate(x_train, axis=0)
    y_train = np.concatenate(y_train, axis=0)

    return x_train, y_train



from __future__ import division, print_function
import numpy as np


class ReccurenceFeaturesExtractor:
    def __init__(self,
                 recurrence_matrix: np.ndarray = None):
        self.recurrence_matrix = recurrence_matrix

    def calculate_DFD(self, number_of_vectors):
        # Calculating the diagonal frequency distribution - P(l)
        diagonal_frequency_distribution = np.zeros(number_of_vectors + 1)
        for i in range(number_of_vectors - 1, -1, -1):
            diagonal_line_length = 0
            for j in range(0, number_of_vectors - i):
                if self.recurrence_matrix[i + j, j] == 1:
                    diagonal_line_length += 1
                    if j == (number_of_vectors - i - 1):
                        diagonal_frequency_distribution[diagonal_line_length] += 1.0
                else:
                    if diagonal_line_length != 0:
                        diagonal_frequency_distribution[diagonal_line_length] += 1.0
                        diagonal_line_length = 0

        for k in range(1, number_of_vectors):
            diagonal_line_length = 0
            for i in range(number_of_vectors - k):
                j = i + k
                if self.recurrence_matrix[i, j] == 1:
                    diagonal_line_length += 1
                    if j == (number_of_vectors - 1):
                        diagonal_frequency_distribution[diagonal_line_length] += 1.0
                else:
                    if diagonal_line_length != 0:
                        diagonal_frequency_distribution[diagonal_line_length] += 1.0
                        diagonal_line_length = 0
        return diagonal_frequency_distribution

    def calculate_VFD(self, number_of_vectors):
        # Calculating the vertical frequency distribution - P(v)
        vertical_frequency_distribution = np.zeros(number_of_vectors + 1)
        for i in range(number_of_vectors):
            vertical_line_length = 0
            for j in range(number_of_vectors):
                if self.recurrence_matrix[i, j] == 1:
                    vertical_line_length += 1
                    if j == (number_of_vectors - 1):
                        vertical_frequency_distribution[vertical_line_length] += 1.0
                else:
                    if vertical_line_length != 0:
                        vertical_frequency_distribution[vertical_line_length] += 1.0
                        vertical_line_length = 0
        return vertical_frequency_distribution

    def calculate_WVFD(self, number_of_vectors):
        # Calculating the white vertical frequency distribution - P(w)
        white_vertical_frequency_distribution = np.zeros(number_of_vectors + 1)
        for i in range(number_of_vectors):
            white_vertical_line_length = 0
            for j in range(number_of_vectors):
                if self.recurrence_matrix[i, j] == 0:
                    white_vertical_line_length += 1
                    if j == (number_of_vectors - 1):
                        white_vertical_frequency_distribution[white_vertical_line_length] += 1.0
                else:
                    if white_vertical_line_length != 0:
                        white_vertical_frequency_distribution[white_vertical_line_length] += 1.0
                        white_vertical_line_length = 0
        return white_vertical_frequency_distribution

    def calculate_EVWL(self, white_vertical_frequency_distribution, MWVL, number_of_vectors):
        longest_white_vertical_line_length = 1
        # Calculating the longest white vertical line length - Wmax
        for w in range(number_of_vectors, 0, -1):
            if white_vertical_frequency_distribution[w] != 0:
                longest_white_vertical_line_length = w
                break

        # Calculating the entropy white vertical lines - Wentr
        sum_white_vertical_frequency_distribution = np.float(
            np.sum(white_vertical_frequency_distribution[MWVL:]))
        entropy_white_vertical_lines = 0
        for w in range(MWVL, number_of_vectors + 1):
            if white_vertical_frequency_distribution[w] != 0:
                entropy_white_vertical_lines += (white_vertical_frequency_distribution[
                                                     w] / sum_white_vertical_frequency_distribution) * np.log(
                    white_vertical_frequency_distribution[w] / sum_white_vertical_frequency_distribution)
        entropy_white_vertical_lines *= -1

        return entropy_white_vertical_lines, longest_white_vertical_line_length

    def recurrence_quantification_analysis(self,
                                           MDL=3,
                                           MVL=3,
                                           MWVL=2):
        # Calculating the number of states - N
        number_of_vectors = self.recurrence_matrix.shape[0]

        # Calculating the recurrence rate - RR
        recurrence_rate = np.float(np.sum(self.recurrence_matrix)) / np.power(number_of_vectors, 2)

        diagonal_frequency_distribution = self.calculate_DFD(
            number_of_vectors=number_of_vectors)

        vertical_frequency_distribution = self.calculate_VFD(number_of_vectors=number_of_vectors)

        white_vertical_frequency_distribution = self.calculate_WVFD(number_of_vectors=number_of_vectors)

        # Calculating the determinism - DET
        numerator = np.sum(
            [l * diagonal_frequency_distribution[l] for l in range(MDL, number_of_vectors)])
        denominator = np.sum([l * diagonal_frequency_distribution[l] for l in range(1, number_of_vectors)])
        determinism = numerator / denominator

        # Calculating the average diagonal line length - L
        numerator = np.sum(
            [l * diagonal_frequency_distribution[l] for l in range(MDL, number_of_vectors)])
        denominator = np.sum(
            [diagonal_frequency_distribution[l] for l in range(MDL, number_of_vectors)])
        average_diagonal_line_length = numerator / denominator
        longest_diagonal_line_length = 1

        # Calculating the longest diagonal line length - Lmax
        for l in range(number_of_vectors - 1, 0, -1):
            if diagonal_frequency_distribution[l] != 0:
                longest_diagonal_line_length = l
                break

        # Calculating the  divergence - DIV
        divergence = 1. / longest_diagonal_line_length

        # Calculating the entropy diagonal lines - Lentr
        sum_diagonal_frequency_distribution = np.float(
            np.sum(diagonal_frequency_distribution[MDL:-1]))
        entropy_diagonal_lines = 0
        for l in range(MDL, number_of_vectors):
            if diagonal_frequency_distribution[l] != 0:
                entropy_diagonal_lines += (diagonal_frequency_distribution[
                                               l] / sum_diagonal_frequency_distribution) * np.log(
                    diagonal_frequency_distribution[l] / sum_diagonal_frequency_distribution)
        entropy_diagonal_lines *= -1

        # Calculating the ratio determinism_recurrence - DET/RR
        ratio_determinism_recurrence_rate = determinism / recurrence_rate

        # Calculating the laminarity - LAM
        numerator = np.sum([v * vertical_frequency_distribution[v] for v in
                            range(MVL, number_of_vectors + 1)])
        denominator = np.sum([v * vertical_frequency_distribution[v] for v in range(1, number_of_vectors + 1)])
        laminarity = numerator / denominator

        # Calculating the average vertical line length - V
        numerator = np.sum([v * vertical_frequency_distribution[v] for v in
                            range(MVL, number_of_vectors + 1)])
        denominator = np.sum(
            [vertical_frequency_distribution[v] for v in range(MVL, number_of_vectors + 1)])
        average_vertical_line_length = numerator / denominator

        longest_vertical_line_length = 1
        # Calculating the longest vertical line length - Vmax
        for v in range(number_of_vectors, 0, -1):
            if vertical_frequency_distribution[v] != 0:
                longest_vertical_line_length = v
                break

        # Calculating the entropy vertical lines - Ventr
        sum_vertical_frequency_distribution = np.float(
            np.sum(vertical_frequency_distribution[MVL:]))
        entropy_vertical_lines = 0
        for v in range(MVL, number_of_vectors + 1):
            if vertical_frequency_distribution[v] != 0:
                entropy_vertical_lines += (vertical_frequency_distribution[
                                               v] / sum_vertical_frequency_distribution) * np.log(
                    vertical_frequency_distribution[v] / sum_vertical_frequency_distribution)
        entropy_vertical_lines *= -1

        # Calculatint the ratio laminarity_determinism - LAM/DET
        ratio_laminarity_determinism = laminarity / determinism

        # Calculating the average white vertical line length - W
        numerator = np.sum([w * white_vertical_frequency_distribution[w] for w in
                            range(MWVL, number_of_vectors + 1)])
        denominator = np.sum([white_vertical_frequency_distribution[w] for w in
                              range(MWVL, number_of_vectors + 1)])
        average_white_vertical_line_length = numerator / denominator

        entropy_white_vertical_lines, longest_white_vertical_line_length = self.calculate_EVWL(
            white_vertical_frequency_distribution=white_vertical_frequency_distribution,
            MWVL=MWVL,
            number_of_vectors=number_of_vectors)

        feature_dict = {
            # 'DFD': diagonal_frequency_distribution,
            #             'VFD': vertical_frequency_distribution,
            #             'WVFD': white_vertical_frequency_distribution,
            'RR': recurrence_rate,
            'DET': determinism,
            'ADLL': average_diagonal_line_length,
            'LDLL': longest_diagonal_line_length,
            'Div': divergence,
            'EDL': entropy_diagonal_lines,
            'Lam': laminarity,
            'AVLL': average_vertical_line_length,
            'LVLL': longest_vertical_line_length,
            'EVL': entropy_vertical_lines,
            'AWLL': average_white_vertical_line_length,
            'LWLL': longest_white_vertical_line_length,
            'EWLL': entropy_white_vertical_lines,
            'RDRR': ratio_determinism_recurrence_rate,
            'RLD': ratio_laminarity_determinism}
        return feature_dict



import torch
import torch.nn as nn
from typing import Optional
from fedot.core.operations.operation_parameters import OperationParameters
from fedot_ind.core.models.nn.modules import Reshape, Flatten, pass_through


class Inception(nn.Module):
    def __init__(self, in_channels,
                 n_filters,
                 kernel_sizes=[9, 19, 39],
                 bottleneck_channels=32,
                 activation=nn.ReLU(),
                 return_indices=False):
        """
        : param in_channels				Number of input channels (input features)
        : param n_filters				Number of filters per convolution layer => out_channels = 4*n_filters
        : param kernel_sizes			List of kernel sizes for each convolution.
                                        Each kernel size must be odd number that meets -> "kernel_size % 2 !=0".
                                        This is nessesery because of padding size.
                                        For correction of kernel_sizes use function "correct_sizes".
        : param bottleneck_channels		Number of output channels in bottleneck.
                                        Bottleneck wont be used if nuber of in_channels is equal to 1.
        : param activation				Activation function for output tensor (nn.ReLU()).
        : param return_indices			Indices are needed only if we want to create decoder with InceptionTranspose with MaxUnpool1d.
        """
        super(Inception, self).__init__()
        self.return_indices = return_indices
        if in_channels > 1:
            self.bottleneck = nn.Conv1d(
                in_channels=in_channels,
                out_channels=bottleneck_channels,
                kernel_size=1,
                stride=1,
                bias=False
            )
        else:
            self.bottleneck = pass_through
            bottleneck_channels = 1

        self.conv_from_bottleneck_1 = nn.Conv1d(
            in_channels=bottleneck_channels,
            out_channels=n_filters,
            kernel_size=kernel_sizes[0],
            stride=1,
            padding=kernel_sizes[0] // 2,
            bias=False
        )
        self.conv_from_bottleneck_2 = nn.Conv1d(
            in_channels=bottleneck_channels,
            out_channels=n_filters,
            kernel_size=kernel_sizes[1],
            stride=1,
            padding=kernel_sizes[1] // 2,
            bias=False
        )
        self.conv_from_bottleneck_3 = nn.Conv1d(
            in_channels=bottleneck_channels,
            out_channels=n_filters,
            kernel_size=kernel_sizes[2],
            stride=1,
            padding=kernel_sizes[2] // 2,
            bias=False
        )
        self.max_pool = nn.MaxPool1d(kernel_size=3, stride=1, padding=1, return_indices=return_indices)
        self.conv_from_maxpool = nn.Conv1d(
            in_channels=in_channels,
            out_channels=n_filters,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=False
        )
        self.batch_norm = nn.BatchNorm1d(num_features=4 * n_filters)
        self.activation = activation

    def forward(self, X):
        # step 1
        Z_bottleneck = self.bottleneck(X)
        if self.return_indices:
            Z_maxpool, indices = self.max_pool(X)
        else:
            Z_maxpool = self.max_pool(X)
        # step 2
        Z1 = self.conv_from_bottleneck_1(Z_bottleneck)
        Z2 = self.conv_from_bottleneck_2(Z_bottleneck)
        Z3 = self.conv_from_bottleneck_3(Z_bottleneck)
        Z4 = self.conv_from_maxpool(Z_maxpool)
        # step 3
        Z = torch.cat([Z1, Z2, Z3, Z4], axis=1)
        Z = self.activation(self.batch_norm(Z))
        if self.return_indices:
            return Z, indices
        else:
            return Z


class InceptionBlock(nn.Module):
    def __init__(self, in_channels, n_filters=32, kernel_sizes=[9, 19, 39], bottleneck_channels=32, use_residual=True,
                 activation=nn.ReLU(), return_indices=False):
        super(InceptionBlock, self).__init__()
        self.use_residual = use_residual
        self.return_indices = return_indices
        self.activation = activation
        self.inception_1 = Inception(
            in_channels=in_channels,
            n_filters=n_filters,
            kernel_sizes=kernel_sizes,
            bottleneck_channels=bottleneck_channels,
            activation=activation,
            return_indices=return_indices
        )
        self.inception_2 = Inception(
            in_channels=4 * n_filters,
            n_filters=n_filters,
            kernel_sizes=kernel_sizes,
            bottleneck_channels=bottleneck_channels,
            activation=activation,
            return_indices=return_indices
        )
        self.inception_3 = Inception(
            in_channels=4 * n_filters,
            n_filters=n_filters,
            kernel_sizes=kernel_sizes,
            bottleneck_channels=bottleneck_channels,
            activation=activation,
            return_indices=return_indices
        )
        if self.use_residual:
            self.residual = nn.Sequential(
                nn.Conv1d(
                    in_channels=in_channels,
                    out_channels=4 * n_filters,
                    kernel_size=1,
                    stride=1,
                    padding=0
                ),
                nn.BatchNorm1d(
                    num_features=4 * n_filters
                )
            )

    def forward(self, X):
        if self.return_indices:
            Z, i1 = self.inception_1(X)
            Z, i2 = self.inception_2(Z)
            Z, i3 = self.inception_3(Z)
        else:
            Z = self.inception_1(X)
            Z = self.inception_2(Z)
            Z = self.inception_3(Z)
        if self.use_residual:
            Z = Z + self.residual(X)
            Z = self.activation(Z)
        if self.return_indices:
            return Z, [i1, i2, i3]
        else:
            return Z


class InceptionTranspose(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_sizes=[9, 19, 39], bottleneck_channels=32,
                 activation=nn.ReLU()):
        """
        : param in_channels				Number of input channels (input features)
        : param n_filters				Number of filters per convolution layer => out_channels = 4*n_filters
        : param kernel_sizes			List of kernel sizes for each convolution.
                                        Each kernel size must be odd number that meets -> "kernel_size % 2 !=0".
                                        This is nessesery because of padding size.
                                        For correction of kernel_sizes use function "correct_sizes".
        : param bottleneck_channels		Number of output channels in bottleneck.
                                        Bottleneck wont be used if nuber of in_channels is equal to 1.
        : param activation				Activation function for output tensor (nn.ReLU()).
        """
        super(InceptionTranspose, self).__init__()
        self.activation = activation
        self.conv_to_bottleneck_1 = nn.ConvTranspose1d(
            in_channels=in_channels,
            out_channels=bottleneck_channels,
            kernel_size=kernel_sizes[0],
            stride=1,
            padding=kernel_sizes[0] // 2,
            bias=False
        )
        self.conv_to_bottleneck_2 = nn.ConvTranspose1d(
            in_channels=in_channels,
            out_channels=bottleneck_channels,
            kernel_size=kernel_sizes[1],
            stride=1,
            padding=kernel_sizes[1] // 2,
            bias=False
        )
        self.conv_to_bottleneck_3 = nn.ConvTranspose1d(
            in_channels=in_channels,
            out_channels=bottleneck_channels,
            kernel_size=kernel_sizes[2],
            stride=1,
            padding=kernel_sizes[2] // 2,
            bias=False
        )
        self.conv_to_maxpool = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=False
        )
        self.max_unpool = nn.MaxUnpool1d(kernel_size=3, stride=1, padding=1)
        self.bottleneck = nn.Conv1d(
            in_channels=3 * bottleneck_channels,
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
            bias=False
        )
        self.batch_norm = nn.BatchNorm1d(num_features=out_channels)

    def forward(self, X, indices):
        Z1 = self.conv_to_bottleneck_1(X)
        Z2 = self.conv_to_bottleneck_2(X)
        Z3 = self.conv_to_bottleneck_3(X)
        Z4 = self.conv_to_maxpool(X)

        Z = torch.cat([Z1, Z2, Z3], axis=1)
        MUP = self.max_unpool(Z4, indices)
        BN = self.bottleneck(Z)
        # another possibility insted of sum BN and MUP is adding 2nd bottleneck transposed convolution

        return self.activation(self.batch_norm(BN + MUP))


class InceptionTransposeBlock(nn.Module):
    def __init__(self, in_channels, out_channels=32, kernel_sizes=[9, 19, 39], bottleneck_channels=32,
                 use_residual=True, activation=nn.ReLU()):
        super(InceptionTransposeBlock, self).__init__()
        self.use_residual = use_residual
        self.activation = activation
        self.inception_1 = InceptionTranspose(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_sizes=kernel_sizes,
            bottleneck_channels=bottleneck_channels,
            activation=activation
        )
        self.inception_2 = InceptionTranspose(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_sizes=kernel_sizes,
            bottleneck_channels=bottleneck_channels,
            activation=activation
        )
        self.inception_3 = InceptionTranspose(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_sizes=kernel_sizes,
            bottleneck_channels=bottleneck_channels,
            activation=activation
        )
        if self.use_residual:
            self.residual = nn.Sequential(
                nn.ConvTranspose1d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=1,
                    stride=1,
                    padding=0
                ),
                nn.BatchNorm1d(
                    num_features=out_channels
                )
            )

    def forward(self, X, indices):
        assert len(indices) == 3
        Z = self.inception_1(X, indices[2])
        Z = self.inception_2(Z, indices[1])
        Z = self.inception_3(Z, indices[0])
        if self.use_residual:
            Z = Z + self.residual(X)
            Z = self.activation(Z)
        return Z


class InceptionTimeNetwork:
    def __init__(self, params: Optional[OperationParameters] = None):
        self.in_channels = params.get('in_channels', 1)
        self.out_channels = params.get('out_channels', 1)
        self.kernel_sizes = params.get('kernel_sizes', [5, 11, 23])
        self.bottleneck_channels = params.get('bottleneck_channels', 32)
        self.use_residual = params.get('use_residual', True)
        self.activation = params.get('activation', nn.ReLU())

    @property
    def network_architecture(self):
        network = nn.Sequential(
            Reshape(out_shape=(1, 160)),
            InceptionBlock(
                in_channels=self.in_channels,
                n_filters=32,
                kernel_sizes=self.kernel_sizes,
                bottleneck_channels=self.bottleneck_channels,
                use_residual=self.use_residual,
                activation=self.activation
            ),
            InceptionBlock(
                in_channels=32 * 4,
                n_filters=32,
                kernel_sizes=self.kernel_sizes,
                bottleneck_channels=self.bottleneck_channels,
                use_residual=self.use_residual,
                activation=self.activation
            ),
            nn.AdaptiveAvgPool1d(output_size=1),
            Flatten(out_features=32 * 4 * 1),
            nn.Linear(in_features=4 * 32 * 1, out_features=4)
        )
        return network




from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import \
    DataOperationImplementation
from fedot.core.repository.dataset_types import DataTypesEnum


class DummyOperation(DataOperationImplementation):

    def fit(self, input_data: InputData):
        pass

    def transform(self, input_data: InputData) -> OutputData:
        predict = self._convert_to_output(input_data, input_data.features, data_type=DataTypesEnum.table)
        return predict



from typing import Union

import numpy as np
import pandas as pd
from sklearn.metrics import (accuracy_score, f1_score,
                             log_loss, mean_absolute_error,
                             mean_absolute_percentage_error,
                             mean_squared_error, mean_squared_log_error,
                             precision_score, r2_score, roc_auc_score)


class ParetoMetrics:
    def __init__(self):
        pass

    def pareto_metric_list(self, costs: Union[list, np.ndarray], maximise: bool = True) -> np.ndarray:
        """ Calculates the pareto front for a list of costs.

        Args:
            costs: list of costs. An (n_points, n_costs) array.
            maximise: flag for maximisation or minimisation.

        Returns:
            A (n_points, ) boolean array, indicating whether each point is Pareto efficient

        """
        is_efficient = np.ones(costs.shape[0], dtype=bool)
        for i, c in enumerate(costs):
            if is_efficient[i]:
                if maximise:
                    is_efficient[is_efficient] = np.any(costs[is_efficient] >= c, axis=1)  # Remove dominated points
                else:
                    is_efficient[is_efficient] = np.any(costs[is_efficient] <= c, axis=1)  # Remove dominated points
        return is_efficient


class QualityMetric:
    def __init__(self, target,
                 predicted_labels,
                 predicted_probs=None,
                 metric_list: list = ('f1', 'roc_auc', 'accuracy', 'logloss', 'precision'),
                 default_value: float = 0.0):
        self.predicted_probs = predicted_probs
        self.predicted_labels = predicted_labels.reshape(-1)
        self.target = target.reshape(-1)
        self.metric_list = metric_list
        self.default_value = default_value

    def metric(self) -> float:
        pass


class RMSE(QualityMetric):
    def metric(self) -> float:
        return mean_squared_error(y_true=self.target, y_pred=self.predicted_labels, squared=False)


class MSE(QualityMetric):
    def metric(self) -> float:
        return mean_squared_error(y_true=self.target, y_pred=self.predicted_labels, squared=True)


class MSLE(QualityMetric):
    def metric(self) -> float:
        return mean_squared_log_error(y_true=self.target, y_pred=self.predicted_labels)


class MAPE(QualityMetric):
    def metric(self) -> float:
        return mean_absolute_percentage_error(y_true=self.target, y_pred=self.predicted_labels)


class F1(QualityMetric):
    def metric(self) -> float:
        target = self.target
        prediction = self.predicted_labels
        self.default_value = 0.0
        n_classes = len(np.unique(target))
        n_classes_pred = len(np.unique(prediction))
        try:
            if n_classes > 2 or n_classes_pred > 2:
                return f1_score(y_true=target, y_pred=prediction, average='weighted')
            else:
                return f1_score(y_true=target, y_pred=prediction, average='binary')
        except ValueError:
            return self.default_value


class MAE(QualityMetric):
    def metric(self) -> float:
        return mean_absolute_error(y_true=self.target, y_pred=self.predicted_labels)


class R2(QualityMetric):
    def metric(self) -> float:
        return r2_score(y_true=self.target, y_pred=self.predicted_labels)


class ROCAUC(QualityMetric):
    def metric(self) -> float:
        n_classes = len(np.unique(self.target))

        self.default_value = 0.5
        if n_classes > 2:
            target = pd.get_dummies(self.target)
            additional_params = {'multi_class': 'ovr', 'average': 'macro'}
            if self.predicted_probs is None:
                prediction = pd.get_dummies(self.predicted_labels)
            else:
                prediction = self.predicted_probs
        else:
            target = self.target
            additional_params = {}
            prediction = self.predicted_labels

        score = roc_auc_score(y_score=prediction, y_true=target, **additional_params)
        score = round(score, 3)

        return score


class Precision(QualityMetric):
    def metric(self) -> float:
        target = self.target
        prediction = self.predicted_labels

        n_classes = np.unique(target)
        if n_classes.shape[0] >= 2:
            additional_params = {'average': 'macro'}
        else:
            additional_params = {}

        score = precision_score(y_pred=prediction, y_true=target, **additional_params)
        score = round(score, 3)
        return score


class Logloss(QualityMetric):
    def metric(self) -> float:
        target = self.target
        prediction = self.predicted_probs
        return log_loss(y_true=target, y_pred=prediction)


class Accuracy(QualityMetric):
    def metric(self) -> float:
        target = self.target
        prediction = self.predicted_labels
        return accuracy_score(y_true=target, y_pred=prediction)



import pandas as pd

from fedot_ind.core.models.BaseExtractor import BaseExtractor
from fedot_ind.core.models.signal.SignalExtractor import SignalExtractor
from fedot_ind.core.models.spectral.SSAExtractor import SSAExtractor
from fedot_ind.core.models.statistical.StatsExtractor import StatsExtractor
from fedot_ind.core.models.topological.TopologicalExtractor import TopologicalExtractor


class EnsembleExtractor(BaseExtractor):
    """Class for performing experiments with ensemble of feature generators.

    Args:
        feature_generator_dict: Dictionary of feature generators consists of
            'generator_name': generator_class pairs.
        list_of_generators: List of feature generators.

    """

    def __init__(self, feature_generator_dict: dict = None,
                 list_of_generators=None, use_cache: bool = False):
        super().__init__(feature_generator_dict)
        self.use_cache = use_cache
        self.list_of_generators = list_of_generators
        self.generator_dict = dict(quantile=StatsExtractor,
                                   window_quantile=StatsExtractor,
                                   wavelet=SignalExtractor,
                                   spectral=SSAExtractor,
                                   spectral_window=SSAExtractor,
                                   topological=TopologicalExtractor,
                                   ensemble=EnsembleExtractor)

    # def get_features(self, ts_frame: pd.DataFrame, dataset_name: str = None, target: np.ndarray = None) -> pd.DataFrame:
    #     return self.ensemble_features(ts_frame, dataset_name)

    def generate_features_from_ts(self, input_data: pd.DataFrame, dataset_name: str = None) -> pd.DataFrame:
        """Extracts features using specified generator and combines them into one feature matrix.

        Args:
            input_data: Dataframe with time series data.
            dataset_name: Dataset name.

        Returns:
            Dataframe with extracted features.

        """
        self.logger.info(f'Extracting features using ensemble of generators: {self.list_of_generators.keys()}')
        features = list()
        for generator_name, generator in self.list_of_generators.items():
            features_df = generator.extract_features(input_data, dataset_name)
            features.append(features_df)

        self.logger.info(f'Features extraction finished')
        return pd.concat(features, axis=1)



import logging

import numpy as np


class MetaFeaturesDetector:

    def __init__(self, train_data, test_data, dataset_name):
        self.train_data = train_data
        self.test_data = test_data
        self.dataset_name = dataset_name
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info(f'Initializing MetaFeaturesDetector for {dataset_name}')

        self.base_metafeatures = ['test_size', 'train_size', 'length', 'number_of_classes', 'type']
        self.extra_metafeatures = []

    def get_base_metafeatures(self):
        train_features_dict = {}
        ts_df_train, ts_target_train = self.train_data[0], self.train_data[1]
        ts_df_test, ts_target_test = self.test_data[0], self.test_data[1]

        train_size = len(ts_target_train)
        test_size = len(ts_target_test)
        num_classes = len(np.unique(ts_target_train))
        length = ts_df_train.shape[1]
        _type = 'custom'
        train_features_dict['train_size'] = train_size
        train_features_dict['test_size'] = test_size
        train_features_dict['number_of_classes'] = num_classes
        train_features_dict['length'] = length
        train_features_dict['type'] = _type

        return train_features_dict

    def get_extra_metafeatures(self):
        pass

    def run(self):
        self.logger.info(f'Running MetaFeaturesDetector for {self.dataset_name}')
        base_metafeatures = self.get_base_metafeatures()
        extra_metafeatures = self.get_extra_metafeatures()
        return {**base_metafeatures, **extra_metafeatures}



from typing import Optional
import numpy as np
from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum
from pymonad.either import Either
from pymonad.list import ListMonad
from tqdm.auto import tqdm

from fedot_ind.core.operation.IndustrialCachableOperation import IndustrialCachableOperationImplementation


class BasisDecompositionImplementation(IndustrialCachableOperationImplementation):
    """
    A class for decomposing data on the abstract basis and evaluating the derivative of the resulting decomposition.
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.n_components = params.get('n_components', 2)
        self.basis = None
        self.data_type = DataTypesEnum.image
        self.min_rank = 1

    def _get_basis(self, data):
        if type(data) == list:
            basis = self._get_multidim_basis(data)
        else:
            basis = self._get_1d_basis(data)
        return basis

    def fit(self, data):
        """Decomposes the given data on the chosen basis.

        Returns:
            np.array: The decomposition of the given data.
        """
        pass

    def _decompose_signal(self):
        pass

    def evaluate_derivative(self, order: int = 1):
        """Evaluates the derivative of the decomposition of the given data.

        Returns:
            np.array: The derivative of the decomposition of the given data.
        """
        pass

    def _transform_one_sample(self, sample: np.array):
        """
            Method for transforming one sample
        """
        pass

    def _get_1d_basis(self, input_data):
        decompose = lambda signal: ListMonad(self._decompose_signal(signal))
        basis = Either.insert(input_data).then(decompose).value[0]
        return basis

    def _transform(self, input_data: InputData) -> np.array:
        """
            Method for transforming all samples
        """
        features = np.array(ListMonad(*input_data.features.tolist()).value)
        v = []
        for series in tqdm(features):
            v.append(self._transform_one_sample(series[~np.isnan(series)]))
        predict = np.array(v)
        return predict

    def _get_multidim_basis(self, data):
        pass

    def _get_multidim_basis(self, input_data):
        decompose = lambda multidim_signal: ListMonad(list(map(self._decompose_signal, multidim_signal)))
        basis = Either.insert(input_data).then(decompose).value[0]
        return basis



import logging
from typing import List, Union
from typing import Optional

import numpy as np
import pandas as pd
from fedot.api.main import Fedot
from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.pipelines.node import PipelineNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_builder import PipelineBuilder
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from golem.core.tuning.simultaneous import SimultaneousTuner

from fedot_ind.api.utils.saver_collections import ResultSaver
from fedot_ind.core.architecture.postprocessing.Analyzer import PerformanceAnalyzer
from fedot_ind.core.architecture.utils.utils import default_path_to_save_results
from fedot_ind.core.repository.initializer_industrial_models import IndustrialModels

np.random.seed(0)


class TimeSeriesClassifierPreset:
    """Class responsible for interaction with Fedot classifier.

    """

    def __init__(self, params: Optional[OperationParameters] = None):
        self.test_data_preprocessed = None
        self.generator_name = 'fedot_preset'
        self.branch_nodes = params.get('branch_nodes', ['data_driven_basis',
                                                        'fourier_basis',
                                                        'wavelet_basis'])

        self.model_params = params.get('model_params')
        self.dataset_name = params.get('dataset')
        self.output_dir = params.get('output_dir', default_path_to_save_results())

        self.saver = ResultSaver(dataset_name=self.dataset_name,
                                 generator_name=self.generator_name,
                                 output_dir=self.output_dir)
        self.logger = logging.getLogger('TimeSeriesClassifier')

        self.prediction_label = None
        self.predictor = None
        self.y_train = None
        self.train_features = None
        self.test_features = None
        self.input_test_data = None

        self.logger.info('TimeSeriesClassifierPreset initialised')


    # TODO: put some datatype
    # TODO: add multidata option
    def _init_input_data(self, X, y):
        input_data = InputData(idx=np.arange(len(X)),
                               features=X.values,
                               target=y.reshape(-1, 1),
                               task=Task(TaskTypesEnum.classification), data_type=DataTypesEnum.table)

        # Multidata option

        # train_data = InputData(idx=np.arange(len(train_data[0])),
        #                        features=np.array(train_data[0].values.tolist()),
        #                        target=train_data[1].reshape(-1, 1),
        #                        task=Task(TaskTypesEnum.classification), data_type=DataTypesEnum.image)

        return input_data

    def _build_pipeline(self):
        pipeline_builder = PipelineBuilder()
        branch_idx = 0
        for node in self.branch_nodes:
            pipeline_builder.add_node(node, branch_idx=branch_idx)
            pipeline_builder.add_node('quantile_extractor', branch_idx=branch_idx)
            branch_idx += 1
        pipeline_builder.join_branches('rf')
        return pipeline_builder.build()

    def _tune_pipeline(self, pipeline: Pipeline, train_data: InputData):
        pipeline_tuner = TunerBuilder(train_data.task) \
            .with_tuner(SimultaneousTuner) \
            .with_metric(ClassificationMetricsEnum.f1) \
            .with_iterations(30) \
            .build(train_data)
        pipeline = pipeline_tuner.tune(pipeline)
        return pipeline

    def fit(self, train_ts_frame,
            train_target: np.ndarray = None,
            **kwargs) -> object:

        with IndustrialModels():
            self.train_data = self._init_input_data(train_ts_frame, train_target)
            self.prerpocessing_pipeline = self._build_pipeline()
            self.prerpocessing_pipeline = self._tune_pipeline(self.prerpocessing_pipeline,
                                                              self.train_data)
            self.prerpocessing_pipeline.fit(self.train_data)

            rf_node = self.prerpocessing_pipeline.nodes[0]
            self.prerpocessing_pipeline.update_node(rf_node, PipelineNode('cat_features'))
            rf_node.nodes_from = []
            rf_node.unfit()
            self.prerpocessing_pipeline.fit(self.train_data)

            train_data_preprocessed = self.prerpocessing_pipeline.root_node.predict(self.train_data)
            train_data_preprocessed.predict = np.squeeze(train_data_preprocessed.predict)

            train_data_preprocessed = InputData(idx=train_data_preprocessed.idx,
                                                features=train_data_preprocessed.predict,
                                                target=train_data_preprocessed.target,
                                                data_type=train_data_preprocessed.data_type,
                                                task=train_data_preprocessed.task)


        metric = 'roc_auc' if train_data_preprocessed.num_classes == 2 else 'f1'
        self.model_params.update({'metric': metric})
        self.predictor = Fedot(**self.model_params)

        self.predictor.fit(train_data_preprocessed)

        return self.predictor

    def predict(self, test_features, test_target) -> dict:
        if self.test_data_preprocessed is None:
            test_data = self._init_input_data(test_features, test_target)
            test_data_preprocessed = self.prerpocessing_pipeline.root_node.predict(test_data)
            test_data_preprocessed.predict = np.squeeze(test_data_preprocessed.predict)
            self.test_data_preprocessed = InputData(idx=test_data_preprocessed.idx,
                                                    features=test_data_preprocessed.predict,
                                                    target=test_data_preprocessed.target,
                                                    data_type=test_data_preprocessed.data_type,
                                                    task=test_data_preprocessed.task)

        self.prediction_label = self.predictor.predict(self.test_data_preprocessed)
        return self.prediction_label

    def predict_proba(self, test_features, test_target) -> dict:
        if self.test_data_preprocessed is None:
            test_data = self._init_input_data(test_features, test_target)
            test_data_preprocessed = self.prerpocessing_pipeline.root_node.predict(test_data)
            self.test_data_preprocessed.predict = np.squeeze(test_data_preprocessed.predict)

        self.prediction_proba = self.predictor.predict_proba(self.test_data_preprocessed)
        return self.prediction_proba

    def get_metrics(self, target: Union[np.ndarray, pd.Series], metric_names: Union[str, List[str]]):
        analyzer = PerformanceAnalyzer()
        return analyzer.calculate_metrics(target=target,
                                          predicted_labels=self.prediction_label,
                                          predicted_probs=self.prediction_proba,
                                          target_metrics=metric_names)

    def save_prediction(self, predicted_data: np.ndarray, kind: str):
        self.saver.save(predicted_data, kind)

    def save_metrics(self, metrics: dict):
        self.saver.save(metrics, 'metrics')



from typing import Tuple, Union

import numpy as np
import pandas as pd
from scipy.signal import argrelextrema, find_peaks


class WindowSizeSelection:
    """Window Size Selection class (WSS).

    This class helps to find appropriate window size in time series to study its features.

    At WSS class we have two group of algorithms. They are Whole-Series-Based (WSB) and subsequence-based (SB).

    Whole-Series-Based (WSB):
        1. 'highest_autocorrelation'
        2. 'dominant_fourier_frequency'

    Subsequence-based (SB):
        1. 'multi_window_finder'
        2. 'summary_statistics_subsequence'

    Note:
        All algorithms has O(n)! Important to set window_max and window_min parameters in case of big time series.

    Args:
        time_series: time series sequences to study.
        window_max: maximum window length to study time series. By default, it is len(time_series)/2.
        window_min: minimum window length to study time series. By default, it is 10.
        wss_algorithm: type of WSS algorithm for your task. By default, it is 'dominant_fourier_frequency'.
            You can choose between:
             'highest_autocorrelation', 'dominant_fourier_frequency',
             'summary_statistics_subsequence' or 'multi_window_finder'.

    Attributes:
        length_ts(int): length of the time_series.

    Reference:
        (c) "Windows Size Selection in Unsupervised Time Series Analytics: A Review and Benchmark. Arik Ermshaus,
    Patrick Schafer, and Ulf Leser. 2022"

    """

    def __init__(self,
                 wss_algorithm: str = 'dominant_fourier_frequency',
                 time_series: Union[list, np.array] = None,
                 window_max: int = None,
                 window_min: int = None):

        self.dict_methods = {'highest_autocorrelation': self.autocorrelation,
                             'multi_window_finder': self.multi_window_finder,
                             'dominant_fourier_frequency': self.dominant_fourier_frequency,
                             'summary_statistics_subsequence': self.summary_statistics_subsequence}

        self.time_series = time_series
        self.window_max = window_max
        self.window_min = window_min
        self.wss_algorithm = wss_algorithm
        self.length_ts = len(time_series)

        if self.window_max is None:
            self.window_max = int(len(self.time_series) / 3)

        if self.window_min is None:
            self.window_min = int(len(self.time_series) / 10)

    def autocorrelation(self) -> Tuple[int, list]:
        """Main function for the highest_autocorrelation (AC) to find an appropriate window size.

        Note:
            self.time_series have to be a list!

        Returns:
            window_size_selected: selected window size for the time series
            distance_scores: score list of DFF metric to understand window size selection

        """
        cutted_ts = []
        list_score = [self.high_ac_metric(self.time_series[i:] + self.time_series[:i], i) \
                      for i in range(self.window_min, self.window_max)]
        selected_size_window, list_score_peaks = self.local_max_search(list_score)
        return selected_size_window, list_score

    def high_ac_metric(self, copy_ts: list, i: int) -> float:
        """Find metric score for the AC.

        Args:
            copy_ts: a list of lagged time series according to window size.
            i: selected window size.

        Returns:
            a_score (float): value which is AC metric for selected window size.

        """
        temp_len = len(copy_ts)
        temp_coef = 1 / (temp_len - i)
        mean_ts = np.mean(copy_ts)
        std_ts = np.std(copy_ts)
        score_list = [(copy_ts[g] - mean_ts) * (self.time_series[g] - mean_ts) / (std_ts ** 2) for g in range(temp_len)]
        a_score = max(score_list) * temp_coef
        return a_score

    def local_max_search(self, score_list: list) -> Tuple[int, list]:
        """Find global max id in time series over score list for AC

        Args:
            score_list: a list of lagged time series according to window size.

        Returns:
            window_size_selected: id for the highest peak in score_list.
            list_score_peaks: list of peaks id in score_list.

        """
        list_probably_peaks = find_peaks(score_list)[0][1:]
        list_score_peaks = [score_list[i] for i in list_probably_peaks]
        max_score = max(list_score_peaks)
        window_size_selected = score_list.index(max_score) + self.window_min
        return window_size_selected, list_score_peaks

    def dominant_fourier_frequency(self) -> Tuple[int, list]:
        """Main function for the dominant_fourier_frequency (DFF) to find an appropriate window size.

        Returns:
            window_size_selected: selected window size for the time series
            distance_scores: score list of DFF metric to understand window size selection

        """
        list_score_k = [self.coeff_metrics(i) for i in range(self.window_min, self.window_max)]
        max_score = max(list_score_k)  # take max value for selected window
        window_size_selected = list_score_k.index(max_score) + self.window_min
        return window_size_selected, list_score_k

    def coeff_metrics(self, temp_size: int) -> float:
        """Find metric score for the DFF.

        Args:
            temp_size: window_size_selected.

        Returns:
            score_coeff: value which is DFF metric distance for selected window size.

        """
        length_n = len(self.time_series)
        temp_list_coeff = [ts * np.exp((-2) * np.pi * (complex(0, -1)) * index * temp_size / length_n) \
                           for index, ts in enumerate(self.time_series)]
        complex_coeff = sum(temp_list_coeff)
        score_coeff = np.sqrt(complex_coeff.real ** 2 + complex_coeff.imag ** 2)
        return score_coeff

    def summary_statistics_subsequence(self, lbound: int = 20, threshold: float = 0.89) -> Tuple[int, list]:
        """Main function for the summary_statistics_subsequence (SuSS) to find an appropriate window size.

        Note:
            This is the fastest implementation - O(log(N))

        Args:
            lbound: first assumption about window_size.
            threshold: normalized anomaly score.

        Returns:
            lbound: selected window size for the time series.
            distance_scores: score list of SuSS score and window_size to understand window size selection.

        """
        ts_fix = np.array(self.time_series)
        ts = (ts_fix - ts_fix.min()) / (ts_fix.max() - ts_fix.min())

        ts_mean = np.mean(ts)
        ts_std = np.std(ts)
        ts_min_max = np.max(ts) - np.min(ts)

        stats = (ts_mean, ts_std, ts_min_max)

        max_score = self.suss_score(ts, 1, stats)
        min_score = self.suss_score(ts, ts.shape[0] - 1, stats)

        exp = 0

        list_score = []
        while True:
            window_size = 2 ** exp
            if window_size < lbound:
                exp += 1
                continue
            score = 1 - (self.suss_score(ts, window_size, stats) - min_score) / (max_score - min_score)
            if score > threshold:
                break
            exp += 1
            list_score.append([score, window_size])
        lbound, ubound = max(lbound, 2 ** (exp - 1)), 2 ** exp + 1
        while lbound <= ubound:
            window_size = int((lbound + ubound) / 2)
            score = 1 - (self.suss_score(ts, window_size, stats) - min_score) / (max_score - min_score)
            list_score.append([score, window_size])
            if score < threshold:
                lbound = window_size + 1
            elif score > threshold:
                ubound = window_size - 1
            else:
                break
        return lbound, list_score

    def suss_score(self, ts: np.array, window_size: int, stats: list) -> float:
        """Find difference between global statistic and statistic of subsequences with different window
           for the SuSS.

        Args:
            ts: normalized numpy time series.
            window_size: temporary selected window size.
            stats: statistic over all normalized time series.

        Returns:
            np.mean(x): not normalized euclidian distance between statistic
                for selected window size and general time series statistic.

        """
        roll = pd.Series(ts).rolling(window_size)
        ts_mean, ts_std, ts_min_max = stats

        roll_mean = roll.mean().to_numpy()[window_size:]
        roll_std = roll.std(ddof=0).to_numpy()[window_size:]
        roll_min = roll.min().to_numpy()[window_size:]
        roll_max = roll.max().to_numpy()[window_size:]

        x = np.array([roll_mean - ts_mean, roll_std - ts_std, (roll_max - roll_min) - ts_min_max])
        x = np.sqrt(np.sum(np.square(x), axis=0)) / np.sqrt(window_size)
        return np.mean(x)

    def multi_window_finder(self) -> Tuple[int, list]:
        """Main multi_window_finder (MWF) function to find an appropriate window size.

        Note:
            In case of 1 global minimum over ts. it is better to try increase max-min boards of ts or change algorithm.

        Returns:
            window_size_selected: selected window size for the time series.
            distance_scores: score list of MWF metric to understand window size selection.

        """
        distance_scores = [self.mwf_metric(i) for i in range(self.window_min, self.window_max)]
        minimum_id_list, id_max = self.top_local_minimum(distance_scores)
        k = 1  # the true first local minimum
        if len(minimum_id_list) < 2:
            k = 0
        window_size_selected = minimum_id_list[k] * 10 + self.window_min + id_max
        return window_size_selected, distance_scores

    def mwf_metric(self, window_selected_temp: int) -> float:
        """Function for MWF method to find metric value for a chosen window size.

        Args:
            window_selected_temp: selected window size for your time series.

        Returns:
            distance_k: value which is the metric distance for selected window size.

        """
        coeff_temp = 1 / window_selected_temp
        m_values = []
        for k in range(self.length_ts - window_selected_temp + 1):
            m_k = [self.time_series[g + k] for g in range(window_selected_temp - 1)]
            m_values.append(coeff_temp * sum(m_k))
        distance_k = sum(np.log10(abs(m_values - np.mean(m_values))))
        return distance_k

    def top_local_minimum(self, distance_scores: list) -> Tuple[int, list]:
        """Find top list of local minimum in scores for MWF method.

        Args:
            distance_scores: list of distance according to MWF metric for selected window sizes

        Returns:
            id_local_minimum_list: list of index where ndarray has local minimum.
            id_max: id where distance_scores has global max value for distance scores.

        """
        id_max = distance_scores.index(max(distance_scores))
        score_temp = distance_scores[id_max:]
        number_windows_temp = len(score_temp) // 10
        scorer_list = [sum(abs(score_temp[i:i + 10] \
                               - np.mean(score_temp[i:i + 10]))) \
                       for i in range(number_windows_temp)]
        id_local_minimum_list = argrelextrema(np.array(scorer_list), np.less)[0]
        return id_local_minimum_list, id_max

    def get_window_size(self) -> Tuple[int, list]:
        """Main function to run WSS class over selected time series

        Note:
            One of the reason of ValueError is that time series size can be equal or smaller than 50.
            In case of it try to initially set window_size min and max.

        Raises:
            ValueError: If `self.window_max` is equal or smaller to `self.window_min`.

        Returns:
            window_size_selected: value which has been chosen as appropriate window size
            list_score: score list for chosen algorithm metric to understand window size selection

        """
        if self.window_max <= self.window_min:
            raise ValueError('Hyperparameters error! self.window_max is equal or smaller to self.window_min')
        else:
            window_size_selected, list_score = self.dict_methods[self.wss_algorithm]()
        return window_size_selected, list_score


class WindowCutter:
    """
    Window cutter.
        input format: dict with "data" and "labels" fields
        output: the same dict but with additional windows_list and labels for it
    """

    def __init__(self, window_len, window_step):
        #super().__init__(name="Window Cutter", operation="window cutting")
        self.window_len = window_len
        self.window_step = window_step
        self.output_window_list = []

    def load_data(self, input_dict: dict) -> None:
        self.input_dict = input_dict

    def get_windows(self) -> dict:
        return self.output_window_list

    def run(self) -> None:
        """
        Cut data to windows
        :return: none
        """
        self.output_window_list = self._cut_ts_to_windows(self.input_dict)

    def _cut_ts_to_windows(self, ts: dict) -> list:
        start_idx = 0
        end_idx = len(ts[list(ts.keys())[0]]) - self.window_len
        temp_windows_list = []
        for i in range(start_idx, end_idx, self.window_step):
            temp_window_dict = {}
            for key in ts.keys():
                temp_window = []
                for j in range(i, i + self.window_len):
                    temp_window.append(ts[key][j])
                temp_window_dict[key] = temp_window
        temp_windows_list.append(temp_window_dict)
        return temp_windows_list



import json
import os

from fedot.core.operations.operation_parameters import OperationParameters


class IndustrialOperationParameters(OperationParameters):
    def from_params(self, operation_type: str, params: OperationParameters):
        return self.from_operation_type(operation_type, **params.to_dict())

    @staticmethod
    def from_operation_type(operation_type: str, **parameters):
        default_parameters = get_default_params(operation_type)
        parameters = {**default_parameters, **parameters}
        return IndustrialOperationParameters(**parameters)


def get_default_params(operation_type: str) -> dict:
    """Gets default params for chosen model name

    :param operation_type: the operation name to choose default parameters for

    :return: default repository parameters for the model name
    """
    with DefaultOperationParamsRepository() as default_params_repo:
        return default_params_repo.get_default_params_for_operation(operation_type)


class DefaultOperationParamsRepository:
    def __init__(self, repository_name: str = 'default_operation_params.json'):
        repo_folder_path = str(os.path.dirname(__file__))
        file = os.path.join('data', repository_name)
        self._repo_path = os.path.join(repo_folder_path, file)
        self._repo = self._initialise_repo()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._repo_path = None

    def _initialise_repo(self) -> dict:
        with open(self._repo_path) as repository_json_file:
            repository_json = json.load(repository_json_file)

        return repository_json

    def get_default_params_for_operation(self, model_name: str) -> dict:
        model_name = model_name.split('/')[0]
        if model_name in self._repo:
            return self._repo[model_name]
        return {}



import logging
import timeit

import pandas as pd


def exception_decorator(exception_return='Problem'):
    def decorate(function):
        def exception_wrapper(*args, **kwargs):
            try:
                function(*args, **kwargs)
            except:
                return print(exception_return)

        return exception_wrapper

    return decorate


def input_data_decorator(func):
    def wrapper(*args, **kwargs):
        data_type = type(kwargs['modelling_results'])
        if data_type != dict:
            if data_type == list:
                modelling_result = [pd.read_csv(element, index_col=0) for element in kwargs['modelling_results']]
                kwargs['modelling_results'] = {str(idx): element for idx, element in enumerate(modelling_result)}
            else:
                kwargs['modelling_results'] = pd.DataFrame(kwargs['modelling_results'])
        result = func(*args, **kwargs)
        return result

    return wrapper


def type_check_decorator(object_type: type, types_list: tuple):
    def type_check_wrapper(function_to_decorate):
        def wrapper(*args, **kwargs):
            if not isinstance(object_type, types_list):
                raise TypeError(f"Unsupported object type. Try one of {str(types_list)}.")
            else:
                function_to_decorate(*args, **kwargs)
            return wrapper

        return type_check_wrapper


def time_it(func):
    def wrapper(*args, **kwargs):
        logger = logging.getLogger('timer')
        start = timeit.default_timer()
        result = func(*args, **kwargs)
        end = timeit.default_timer()
        logger.info(f'Time spent - {round((end - start), 2)} sec')
        return result

    return wrapper


def dataframe_adapter(func):
    def wrapper(*args, **kwargs):
        if not isinstance(kwargs['ts_frame'], pd.DataFrame):
            if isinstance(kwargs['ts_frame'], list):
                kwargs['ts_frame'] = pd.DataFrame(kwargs['ts_frame'])
            else:
                raise TypeError(f'Unsupported input type for {args[0].__class__.__name__}')
        return func(*args, **kwargs)
    return wrapper




from typing import Optional
import numpy as np
from fedot.core.operations.operation_parameters import OperationParameters
from fedot_ind.core.operation.transformation.basis.abstract_basis import BasisDecompositionImplementation


class FourierBasisImplementation(BasisDecompositionImplementation):
    """A class for decomposing data on the Fourier basis and evaluating the derivative of the resulting decomposition.
        Example:
            ts1 = np.random.rand(200)
            ts2 = np.random.rand(200)
            ts = [ts1, ts2]
            bss = FourierBasisImplementation({'spectrum_type': 'real'})
            basis_multi = bss._transform(ts)
            basis_1d = bss._transform(ts1)

    Attributes:
        data (np.array): The array of data to be decomposed.
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.spectrum_type = params.get('spectrum_type')
        self.threshold = params.get('threshold')
        self.basis = None

    def low_pass(self, input_data):
        fourier_coef = np.fft.rfft(input_data)
        frequencies = np.fft.rfftfreq(input_data.size, d=2e-3 / input_data.size)
        fourier_coef[frequencies > self.threshold] = 0
        return np.fft.irfft(fourier_coef)

    def _decompose_signal(self, input_data):
        spectrum = np.fft.fft(input_data)
        if self.spectrum_type == 'imaginary':
            spectrum = spectrum.imag
        elif self.spectrum_type == 'smoothed':
            spectrum = self.low_pass(input_data)
        else:
            spectrum = spectrum.real
        return spectrum

    def _transform_one_sample(self, series: np.array):
        return self._get_basis(series)

    def evaluate_derivative(self, order):
        """Evaluates the derivative of the Fourier decomposition of the given data.

        Returns:
            np.array: The derivative of the Fourier decomposition of the given data.
        """
        return np.fft.ifft(1j * np.arange(len(self.data_range)) * self.decomposed)



from collections import OrderedDict
from typing import Dict, Optional, Tuple, List

import torch
from torch import Tensor
from torch.linalg import vector_norm
from torch.nn import Conv2d

from fedot_ind.core.models.cnn.sfp_models import SFP_MODELS


def zerolize_filters(conv: Conv2d, pruning_ratio: float) -> None:
    """Zerolize filters of convolutional layer to the pruning_ratio (in-place).

    Args:
        conv: The optimizable layer.
        pruning_ratio: pruning hyperparameter must be in the range (0, 1),
            percentage of zerolized filters.
    """
    assert 0 < pruning_ratio < 1, "pruning_ratio must be in the range (0, 1)"
    filter_pruned_num = int(conv.weight.size()[0] * pruning_ratio)
    filter_norms = vector_norm(conv.weight, dim=(1, 2, 3))
    _, indices = filter_norms.sort()
    with torch.no_grad():
        conv.weight[indices[:filter_pruned_num]] = 0


def _check_nonzero_filters(weight: Tensor) -> Tensor:
    """Returns indices of zero filters."""
    filters = torch.count_nonzero(weight, dim=(1, 2, 3))
    indices = torch.flatten(torch.nonzero(filters))
    return indices


def _prune_filters(
        weight: Tensor,
        saving_filters: Optional[Tensor] = None,
        saving_channels: Optional[Tensor] = None,
) -> Tensor:
    """Prune filters and channels of convolutional layer.

    Args:
        weight: Weight matrix.
        saving_filters: Indexes of filters to be saved.
            If ``None`` all filters to be saved.
        saving_channels: Indexes of channels to be saved.
            If ``None`` all channels to be saved.
    """
    if saving_filters is not None:
        weight = weight[saving_filters].clone()
    if saving_channels is not None:
        weight = weight[:, saving_channels].clone()
    return weight


def _prune_batchnorm(bn: Dict, saving_channels: Tensor) -> Dict[str, Tensor]:
    """Prune BatchNorm2d.

    Args:
        bn: Dictionary with batchnorm params.
        saving_channels: Indexes of channels to be saved.
            If ``None`` all channels to be saved.
    """
    bn['weight'] = bn['weight'][saving_channels].clone()
    bn['bias'] = bn['bias'][saving_channels].clone()
    bn['running_mean'] = bn['running_mean'][saving_channels].clone()
    bn['running_var'] = bn['running_var'][saving_channels].clone()
    return bn


def _index_union(x: Tensor, y: Tensor) -> Tensor:
    """Returns the union of x and y"""
    x = set(x.tolist())
    y = set(y.tolist())
    xy = x | y
    return torch.tensor(list(xy))


def _indexes_of_tensor_values(tensor: Tensor, values: Tensor) -> Tensor:
    """Returns the indexes of the values in the input tensor."""
    indexes = []
    tensor = tensor.tolist()
    for value in values.tolist():
        indexes.append(tensor.index(value))
    return torch.tensor(indexes)


def _parse_sd(state_dict: OrderedDict) -> OrderedDict:
    """Parses state_dict to nested dictionaries."""
    parsed_sd = OrderedDict()
    for k, v in state_dict.items():
        _parse_param(k.split('.'), v, parsed_sd)
    return parsed_sd


def _parse_param(param, value, dictionary) -> None:
    """Parses value from state_dict to nested dictionaries."""
    if len(param) > 1:
        dictionary.setdefault(param[0], OrderedDict())
        _parse_param(param[1:], value, dictionary[param[0]])
    else:
        dictionary[param[0]] = value


def _collect_sd(parsed_state_dict) -> OrderedDict:
    """Collect state_dict from nested dictionaries."""
    state_dict = OrderedDict()
    keys, values = _collect_param(parsed_state_dict)
    for k, v in zip(keys, values):
        key = '.'.join(k)
        state_dict[key] = v
    return state_dict


def _collect_param(dictionary) -> Tuple:
    """Collect value from nested dictionaries."""
    if isinstance(dictionary, OrderedDict):
        all_keys = []
        all_values = []
        for k, v in dictionary.items():
            keys, values = _collect_param(v)
            for key in keys:
                key.insert(0, k)
            all_values.extend(values)
            all_keys.extend(keys)
        return all_keys, all_values
    else:
        return [[]], [dictionary]


def _prune_resnet_block(block: Dict, input_channels: Tensor) -> Tuple[Tensor, Tensor]:
    """Prune block of ResNet"""
    channels = input_channels
    downsample_channels = input_channels
    keys = list(block.keys())
    if 'downsample' in keys:
        filters = _check_nonzero_filters(block['downsample']['0']['weight'])
        block['downsample']['0']['weight'] = _prune_filters(
            weight=block['downsample']['0']['weight'],
            saving_filters=filters,
            saving_channels=downsample_channels
        )
        downsample_channels = filters
        block['downsample']['1'] = _prune_batchnorm(
            bn=block['downsample']['1'],
            saving_channels=downsample_channels
        )
        keys.remove('downsample')
    final_conv = keys[-2]
    final_bn = keys[-1]
    keys = keys[:-2]
    for key in keys:
        if key.startswith('conv'):
            filters = _check_nonzero_filters(block[key]['weight'])
            block[key]['weight'] = _prune_filters(
                weight=block[key]['weight'],
                saving_filters=filters,
                saving_channels=channels
            )
            channels = filters
        elif key.startswith('bn'):
            block[key] = _prune_batchnorm(bn=block[key], saving_channels=channels)
    filters = _check_nonzero_filters(block[final_conv]['weight'])
    filters = _index_union(filters, downsample_channels)
    block[final_conv]['weight'] = _prune_filters(
        weight=block[final_conv]['weight'],
        saving_filters=filters,
        saving_channels=channels,
    )
    channels = filters
    block[final_bn] = _prune_batchnorm(bn=block[final_bn], saving_channels=channels)
    block['indexes'] = _indexes_of_tensor_values(channels, downsample_channels)
    return channels, _indexes_of_tensor_values(channels, downsample_channels)


def prune_resnet_state_dict(
        state_dict: OrderedDict
) -> Tuple[OrderedDict, Dict[str, List[int]], Dict[str, List[int]]]:
    """Prune state_dict of ResNet

    Args:
        state_dict: ``state_dict`` of ResNet model.

    Returns:
        Tuple(state_dict, input_channels, output_channels).
    """
    input_size = {'layer1': [], 'layer2': [], 'layer3': [], 'layer4': []}
    output_size = {'layer1': [], 'layer2': [], 'layer3': [], 'layer4': []}
    sd = _parse_sd(state_dict)
    filters = _check_nonzero_filters(sd['conv1']['weight'])
    sd['conv1']['weight'] = _prune_filters(
        weight=sd['conv1']['weight'], saving_filters=filters
    )
    channels = filters
    sd['bn1'] = _prune_batchnorm(bn=sd['bn1'], saving_channels=channels)
    for layer in ['layer1', 'layer2', 'layer3', 'layer4']:
        for k, v in sd[layer].items():
            input_size[layer].append(channels.size()[0])
            channels, index = _prune_resnet_block(block=v, input_channels=channels)
            output_size[layer].append(channels.size()[0])
    sd['fc']['weight'] = sd['fc']['weight'][:, channels].clone()
    sd = _collect_sd(sd)
    return sd, input_size, output_size


def load_sfp_resnet_model(
        model_name: str,
        num_classes: int,
        state_dict_path: str,
        pruning_ratio: float,
) -> torch.nn.Module:
    """Loads SFP state_dict to model.

    Args:
        model_name: Name of the model.
        num_classes: Number of classes.
        state_dict_path: Path to state_dict file.
        pruning_ratio: pruning hyperparameter used in training.
    """
    if model_name not in SFP_MODELS.keys():
        raise ValueError(
            f"model_name must be one of {SFP_MODELS.keys()}, but got '{model_name}'"
        )
    state_dict = torch.load(state_dict_path, map_location='cpu')
    input_size = {'layer1': [], 'layer2': [], 'layer3': [], 'layer4': []}
    output_size = {'layer1': [], 'layer2': [], 'layer3': [], 'layer4': []}
    last_layer = ''
    for k, v in state_dict.items():
        k = k.split('.')
        if len(k) > 3 and k[2] == 'conv1':
            input_size[k[0]].append(v.size()[1])
            if last_layer in output_size.keys():
                output_size[last_layer].append(v.size()[1])
            last_layer = k[0]
    output_size['layer4'].append(state_dict['fc.weight'].size()[1])

    model = SFP_MODELS[model_name](
        num_classes=num_classes,
        input_size=input_size,
        output_size=output_size,
        pruning_ratio=pruning_ratio
    )
    model.load_state_dict(state_dict)
    return model



import numpy as np
from typing import Tuple, TypeVar
from fedot_ind.core.operation.transformation.basis.abstract_basis import BasisDecompositionImplementation

class_type = TypeVar("T", bound="PowerBasis")


class PowerBasis(BasisDecompositionImplementation):
    """Power basis.
    Basis formed by powers of the argument :math:`t`:
    .. math::
        1, t, t^2, t^3...
    Attributes:
        data_range: a tuple of length 2 containing the initial and
            end values of the interval over which the basis can be evaluated.
        n_components: number of functions in the basis.
    """

    def _get_basis(self, n_components: int = None):
        self.basis = np.arange(self.n_components)

        return self.basis

    def fit(self, data):
        projected_data = np.power.outer(data, self.basis)
        return projected_data

    def evaluate_derivative(self: class_type,
                            coefs: np.array,
                            order: int = 1) -> Tuple[class_type, np.array]:
        basis = type(self)(
            domain_range=self.domain_range,
            n_basis=self.n_basis - order,
        )
        derivative_coefs = np.array([np.polyder(x[::-1], order)[::-1] for x in coefs])

        return basis, derivative_coefs





import os

import pandas as pd


class ReporterTSC:
    def __init__(self, path_to_save: str = None):
        self._report = []
        self.path_to_save = path_to_save

    @property
    def report(self):
        return self._report

    def add_to_report(self, dataset_name, generator, launch_num, result):
        self._report.append({'dataset': dataset_name,
                             'generator': generator,
                             'launch_num': launch_num,
                             'result': result})

    def get_summary(self):
        report_df = pd.DataFrame(self.report)
        report_path = os.path.join(self.path_to_save, 'summary.csv')
        report_df.to_csv(report_path, index=False)
        return report_df


if __name__ == '__main__':
    reporter = ReporterTSC()
    reporter.add('beef', 'quantile', 1, 'good')
    reporter.add('beef', 'quantile', 2, 'bad')
    reporter.add('beef', 'wavelet', 1, 'good')
    reporter.add('beef', 'wavelet', 2, 'bad')
    reporter.add('lightning', 'quantile', 1, 'good')
    reporter.add('lightning', 'quantile', 2, 'bad')

    print(reporter.get('dataset_name'))



import os
import shutil
import time
from typing import Dict, List, Optional, Set, Type, Union

import numpy as np
import torch
from PIL import Image
from sklearn.metrics import accuracy_score, roc_auc_score
from sklearn.metrics import f1_score, precision_recall_fscore_support
from torch.nn.functional import softmax
from torch.optim.lr_scheduler import StepLR
from torch.utils.data import DataLoader, Dataset
from torch.utils.tensorboard import SummaryWriter
from torchmetrics.detection.mean_ap import MeanAveragePrecision
from torchvision.models.detection import fasterrcnn_resnet50_fpn
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from tqdm import tqdm

from fedot_ind.core.models.cnn.classification_models import CLF_MODELS
from fedot_ind.core.operation.optimization.ModelStructure import OPTIMIZATIONS


def _parameter_value_check(parameter: str, value: str, valid_values: Set[str]) -> None:
    """Checks if the parameter is in the set of allowed.

    Args:
        parameter: Name of the checked parameter.
        value: Value of the checked parameter.
        valid_values: Set of the valid parameter values.
    """
    if value not in valid_values:
        raise ValueError(
            f"{parameter} must be one of {valid_values}, but got {parameter}='{value}'"
        )


class _GeneralizedExperimenter:
    """Generalized class for working with models.

    Args:
        model: Trainable model.
        optimizable_module_name: Name of the module whose structure will be optimized.
        train_ds: Train dataset.
        val_ds: Validation dataset.
        num_classes: Number of classes in the dataset.
        dataloader_params: Parameter dictionary passed to dataloaders.
        name: Description of the experiment.
        models_path: Path to folder for saving models.
        summary_path: Path to folder for writing experiment summary info.
        summary_per_class: If ``True``, calculates the metrics for each class.
        weights: Path to the model state_dict to load weights (default: ``None``).
        metric: Target metric by which models are compared.
        gpu: If ``True``, uses GPU (default: ``True``).
    """

    def __init__(
        self,
        model: torch.nn.Module,
        optimizable_module_name: str,
        train_ds: Dataset,
        val_ds: Dataset,
        num_classes: int,
        dataloader_params: Dict,
        name: str,
        models_path: str,
        summary_path: str,
        summary_per_class: bool,
        metric: str,
        weights: Optional[str] = None,
        gpu: bool = True,
    ) -> None:
        self.model = model
        if weights is not None:
            self.model.load_state_dict(torch.load(weights))
        self.optimizable_module_name = optimizable_module_name
        print(f"Default size: {self.size_of_model():.2f} MB")
        self.num_classes = num_classes

        self.train_dl = DataLoader(dataset=train_ds, shuffle=True, **dataloader_params)
        self.val_dl = DataLoader(dataset=val_ds, shuffle=False, **dataloader_params)

        self.name = name
        self.models_path = models_path
        self.summary_path = summary_path
        self.summary_per_class = summary_per_class
        self.device = torch.device('cuda' if gpu else 'cpu')
        self.best_score = 0
        self.metric = metric
        self.structure_optimization = None

    def save_model(self, name: str = 'trained') -> None:
        """Save all model.

        Args:
            name: File name (default: 'trained').
        """
        dir_path = os.path.join(self.models_path, self.name)
        model_name = f"{name}.model.pt"
        file_path = os.path.join(dir_path, model_name)
        os.makedirs(dir_path, exist_ok=True)
        try:
            torch.save(self.model, file_path)
        except Exception:
            torch.save(self.model, model_name)
            shutil.move(model_name, dir_path)
        print("Model saved.")

    def save_model_state_dict(self, name: str = 'trained') -> None:
        """Save model state_dict.

        Args:
            name: File name (default: 'trained').
        """
        dir_path = os.path.join(self.models_path, self.name)
        model_name = f"{name}.sd.pt"
        file_path = os.path.join(dir_path, model_name)
        os.makedirs(dir_path, exist_ok=True)
        try:
            torch.save(self.model.state_dict(), file_path)
        except Exception:
            torch.save(self.model.state_dict(), model_name)
            shutil.move(model_name, dir_path)
        print("Model state dict saved.")

    def load_model_state_dict(self, name: str = 'trained') -> None:
        """Load model state_dict to ``self.model``

        Args:
            name: File name (default: 'trained').
        """
        file_path = os.path.join(self.models_path, self.name, f"{name}.sd.pt")
        if os.path.exists(file_path):
            self.model.load_state_dict(torch.load(file_path))
            self.model.to(self.device)
            print("Model state dict loaded.")
        else:
            print(f"File '{file_path}' does not exist.")

    def load_model(self, name: str = 'trained') -> None:
        """Load model to ``self.model``.

        Args:
            name: File name (default: 'trained').
        """
        file_path = os.path.join(self.models_path, self.name, f"{name}.model.pt")
        if os.path.exists(file_path):
            self.model = torch.load(file_path)
            self.model.to(self.device)
            print("Model loaded.")
        else:
            print(f"File '{file_path}' does not exist.")

    def size_of_model(self) -> float:
        """Returns size of model in Mb."""
        param_size = 0
        for param in self.model.parameters():
            param_size += param.nelement() * param.element_size()
        buffer_size = 0
        for buffer in self.model.buffers():
            buffer_size += buffer.nelement() * buffer.element_size()
        return (param_size + buffer_size) / 1e6

    def number_of_params(self) -> int:
        """Returns number of model parameters."""
        return sum(p.numel() for p in self.model.parameters())

    def _read_image(self, image_path: str):
        image = Image.open(image_path).convert('RGB')
        return self.val_dl.dataset.transform(image)

    def train_loop(self) -> Dict[str, float]:
        """Have to implement the training method of the model and return train_scores."""
        return {}

    def val_loop(self) -> Dict[str, float]:
        """Have to implement the validation method of the model and return val_scores."""
        return {}

    def _predict_on_image(self, image: torch.Tensor, proba: bool):
        """Have to implement the prediction method on single image."""
        pass

    def __predict_from_folder(self, image_folder: str, proba: bool) -> Dict:
        predicts = {}
        for image_name in tqdm(os.listdir(image_folder)):
            image_path = os.path.join(image_folder, image_name)
            image = self._read_image(image_path)
            predicts[image_name] = self._predict_on_image(image=image, proba=proba)
        return predicts

    def __predict_from_array(self, image_folder: np.ndarray, proba: bool) -> Dict:
        predicts = {}
        array_iter = enumerate(image_folder)
        for idx, array in tqdm(array_iter):
            image = torch.from_numpy(array).float()
            predicts[str(idx)] = self._predict_on_image(image=image, proba=proba)
        return predicts

    def _predict(self, image_folder: Union[str, np.ndarray], proba: bool) -> Dict:
        """Generalized prediction method of the model."""
        if type(image_folder) == str:
            return self.__predict_from_folder(image_folder, proba)
        else:
            return self.__predict_from_array(image_folder, proba)

    def predict(self, image_folder: Union[str, np.ndarray]) -> Dict:
        """Computes predictions for images in image_folder.

        Args:
            image_folder: Image folder path."""
        return self._predict(image_folder=image_folder, proba=False)

    def predict_proba(self, image_folder: Union[str, np.ndarray]) -> Dict:
        """Computes probability predictions for images in image_folder.

        Args:
            image_folder: Image folder path.
        """
        return self._predict(image_folder=image_folder, proba=True)

    def get_optimizable_module(self) -> torch.nn.Module:
        """Returns the module for optimization applying."""
        return self.model

    def write_scores(
            self,
            writer: SummaryWriter,
            phase: str,
            scores: Dict[str, float],
            x: int,
    ):
        """Write scores from dictionary by SummaryWriter.

        Args:
            writer: SummaryWriter object for writing scores.
            phase: Experiment phase for grouping records, e.g. 'train'.
            scores: Dictionary {metric_name: value}.
            x: The independent variable.
        """
        for key, score in scores.items():
            writer.add_scalar(f"{phase}/{key}", score, x)

    def fit(self, num_epochs: int) -> None:
        """Run optimization experiment.

        Args:
            num_epochs: Number of epochs.
        """
        writer = SummaryWriter(os.path.join(self.summary_path, self.name, 'train'))
        print(f"{self.name}, using device: {self.device}")
        for epoch in range(1, num_epochs + 1):
            print(f"Epoch {epoch}")
            train_scores = self.train_loop()
            self.write_scores(writer, 'train', train_scores, epoch)
            self.structure_optimization.optimize_during_training()
            val_scores = self.val_loop()
            self.write_scores(writer, 'val', val_scores, epoch)
            if val_scores[self.metric] > self.best_score:
                self.best_score = val_scores[self.metric]
                print(f'Best score - {self.best_score}')
                self.save_model_state_dict()
        self.structure_optimization.final_optimize()
        writer.close()

    def finetune(self, num_epochs: int, name: str = 'fine-tuning') -> None:
        """Run fine-tuning

        Args:
            num_epochs: Number of epochs.
            name: Description of the experiment (default: ``'fine-tuning'``).
        """
        best_score = 0
        writer = SummaryWriter(os.path.join(self.summary_path, self.name, name))
        print(f"{self.name}/{name}, using device: {self.device}")
        for epoch in range(1, num_epochs + 1):
            print(f"Fine-tuning epoch {epoch}")
            train_scores = self.train_loop()
            self.write_scores(writer, 'fine-tuning_train', train_scores, epoch)
            val_scores = self.val_loop()
            self.write_scores(writer, 'fine-tuning_val', val_scores, epoch)
            if val_scores[self.metric] > best_score:
                best_score = val_scores[self.metric]
                self.save_model_state_dict(name=name)
        self.load_model_state_dict(name=name)
        writer.close()


class ClassificationExperimenter(_GeneralizedExperimenter):
    """Class for working with classification models.

    Args:
        dataset_name: Name of dataset.
        train_dataset: Train dataset.
        val_dataset: Validation dataset.
        num_classes: Number of classes in the dataset.
        dataloader_params: Parameter dictionary passed to dataloaders.
        model: Name of model.
        model_params: Parameter dictionary passed to model initialization.
        models_saving_path: Path to folder for saving models.
        optimizer: Model optimizer, e.g. ``torch.optim.Adam``.
        optimizer_params: Parameter dictionary passed to optimizer initialization.
        target_loss: Loss function applied to model output,
            e.g. ``torch.nn.CrossEntropyLoss``.
        loss_params: Parameter dictionary passed to loss initialization.
        structure_optimization: Structure optimizer, e.g. ``SVDOptimization``.
        structure_optimization_params: Parameter dictionary passed to structure
            optimization initialization.
        metric: Target metric by which models are compared. May be ``'f1'``,
            ``'accuracy'``, ``'precision'``, ``'recall'`` or ``'roc_auc'``
            (default: ``'f1'``).
        summary_path: Path to folder for writing experiment summary info
            (default: ``'runs'``).
        summary_per_class: If ``True``, calculates the metrics for each class
            (default ``False``).
        weights: Path to the model state_dict to load weights (default: ``None``).
        prefix: An explanatory string added to the name of the experiment.
        gpu: If ``True``, uses GPU (default: ``True``).

        Raises:
            ValueError: If ``model``, ``structure_optimization``, ``metric``
                or ``dataset_name`` not in valid values.
    """

    def __init__(
        self,
        dataset_name: str,
        train_dataset: Dataset,
        val_dataset: Dataset,
        num_classes: int,
        dataloader_params: Dict,
        model: str,
        model_params: Dict,
        models_saving_path: str,
        optimizer: Type[torch.optim.Optimizer],
        optimizer_params: Dict,
        target_loss: Type[torch.nn.Module],
        loss_params: Dict,
        structure_optimization: str,
        structure_optimization_params: Dict,
        metric: str = 'f1',
        summary_path: str = 'runs',
        summary_per_class: bool = False,
        weights: Optional[str] = None,
        prefix: str = '',
        gpu: bool = True,
    ) -> None:

        _parameter_value_check(
            parameter='model', value=model, valid_values=set(CLF_MODELS.keys())
        )
        _parameter_value_check(
            parameter='structure_optimization',
            value=structure_optimization,
            valid_values=set(OPTIMIZATIONS.keys()),
        )
        _parameter_value_check(
            parameter='metric',
            value=metric,
            valid_values={'f1', 'accuracy', 'precision', 'recall', 'roc_auc'},
        )

        super().__init__(
            model=CLF_MODELS[model](num_classes=num_classes, **model_params),
            optimizable_module_name=model,
            train_ds=train_dataset,
            val_ds=val_dataset,
            num_classes=num_classes,
            dataloader_params=dataloader_params,
            name=f"{dataset_name}/{prefix}{model}",
            models_path=models_saving_path,
            summary_path=summary_path,
            summary_per_class=summary_per_class,
            metric=metric,
            gpu=gpu,
        )

        self.structure_optimization = OPTIMIZATIONS[structure_optimization](
            experimenter=self, **structure_optimization_params
        )
        self.target_loss = target_loss(**loss_params)
        self.model.to(self.device)
        self.optimizer = optimizer(self.model.parameters(), **optimizer_params)

    def train_loop(self) -> (Dict[str, float]):
        """Training method of the model.

        Returns:
            Dictionary {metric_name: value}.
        """
        self.model.train()
        train_scores = {'accuracy': 0, 'loss': 0}
        for key in self.structure_optimization.losses:
            train_scores[key] = 0
        for x, y in tqdm(self.train_dl):
            x = x.to(self.device)
            y = y.to(self.device)
            pred = self.model(x)
            loss = self.target_loss(pred, y)
            train_scores['loss'] += loss.item()
            train_scores['accuracy'] += (
                (pred.argmax(1) == y).type(torch.float).mean().item()
            )
            for key, loss_fn in self.structure_optimization.losses.items():
                opt_loss = loss_fn(self.get_optimizable_module())
                loss += opt_loss
                train_scores[key] += opt_loss.item()
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

        for key in train_scores:
            train_scores[key] /= len(self.train_dl)
        return train_scores

    def val_loop(self) -> Dict[Union[str, int], float]:
        """Validation method of the model. Returns val_scores

        Returns:
            Dictionary {metric_name: value}.
        """
        self.model.eval()
        val_loss = 0
        y_true = []
        y_pred = []
        y_score = []
        start = time.time()
        with torch.no_grad():
            for x, y in tqdm(self.val_dl):
                y_true.extend(y)
                x = x.to(self.device)
                y = y.to(self.device)
                pred = self.model(x)
                y_pred.extend(pred.cpu().argmax(1))
                y_score.extend(softmax(pred, dim=1).tolist())
                val_loss += self.target_loss(pred, y).item()
        total_time = time.time() - start
        precision, recall, f1, _ = precision_recall_fscore_support(
            y_true, y_pred, average='macro'
        )
        val_scores = {
            'loss': val_loss / len(self.val_dl),
            'inference_time': total_time / len(self.val_dl),
            'accuracy': accuracy_score(y_true, y_pred),
            'precision': precision,
            'recall': recall,
            'f1': f1,
            'roc_auc': roc_auc_score(y_true, y_score, multi_class='ovo'),
        }
        if self.summary_per_class:
            f1s = f1_score(y_true, y_pred, average=None)
            val_scores.update({f'f1_class{i}': s for i, s in enumerate(f1s)})
        return val_scores

    def _predict_on_image(self, image: torch.Tensor, proba: bool) -> Union[List, int]:
        """Returns prediction for image."""
        self.model.eval()
        with torch.no_grad():

            try:
                pred = self.model(image)
            except Exception:
                image = image.unsqueeze_(0).to(self.device)
                pred = self.model(image)

            if proba:
                pred = softmax(pred, dim=1).cpu().detach().tolist()[0]
            else:
                pred = pred.argmax(1).cpu().detach().item()
        return pred


class FasterRCNNExperimenter(_GeneralizedExperimenter):
    """Class for working with Faster R-CNN model.

        Args:
        dataset_name: Name of dataset.
        train_dataset: Train dataset.
        val_dataset: Validation dataset.
        num_classes: Number of classes in the dataset.
        dataloader_params: Parameter dictionary passed to dataloaders.
        model_params: Parameter dictionary passed to ``fasterrcnn_resnet50_fpn``.
        models_saving_path: Path to folder for saving models.
        optimizer: Model optimizer, e.g. ``torch.optim.SGD``.
        optimizer_params: Parameter dictionary passed to optimizer initialization.
        scheduler_params: Parameter dictionary passed to ``StepLR`` initialization.
        structure_optimization: Structure optimizer, e.g. ``SVDOptimization``.
        structure_optimization_params: Parameter dictionary passed to structure
            optimization initialization.
        metric: Target metric by which models are compared. May be ``'map'``,
            ``'map_50'`` or ``'map_75'`` (default: ``'map_50'``).
        summary_path: Path to folder for writing experiment summary info
            (default: ``'runs'``).
        summary_per_class: If ``True``, calculates the metrics for each class
            (default ``False``).
        gpu: If ``True``, uses GPU (default: ``True``).

        Raises:
            ValueError: If ``model``, ``structure_optimization``, ``metric``
                or ``dataset_name`` not in valid values.
    """

    def __init__(
        self,
        dataset_name: str,
        train_dataset: Dataset,
        val_dataset: Dataset,
        num_classes: int,
        dataloader_params: Dict,
        model_params: Dict,
        models_saving_path: str,
        optimizer: Type[torch.optim.Optimizer],
        optimizer_params: Dict,
        scheduler_params: Dict,
        structure_optimization: str,
        structure_optimization_params: Dict,
        metric: str = 'map_50',
        summary_path: str = 'runs',
        summary_per_class: bool = False,
        weights: Optional[str] = None,
        prefix: str = '',
        gpu: bool = True,
    ) -> None:

        _parameter_value_check(
            parameter='structure_optimization',
            value=structure_optimization,
            valid_values=set(OPTIMIZATIONS.keys()),
        )
        _parameter_value_check(
            parameter='metric',
            value=metric,
            valid_values={'map', 'map_50', 'map_75'},
        )

        model = fasterrcnn_resnet50_fpn(**model_params)
        in_features = model.roi_heads.box_predictor.cls_score.in_features
        model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes)

        super().__init__(
            model=model,
            optimizable_module_name="ResNet50",
            train_ds=train_dataset,
            val_ds=val_dataset,
            num_classes=num_classes,
            dataloader_params=dataloader_params,
            name=f"{dataset_name}/{prefix}FasterR-CNN/ResNet50",
            models_path=models_saving_path,
            summary_path=summary_path,
            summary_per_class=summary_per_class,
            metric=metric,
            weights=weights,
            gpu=gpu,
        )
        self.structure_optimization = OPTIMIZATIONS[structure_optimization](
            experimenter=self, **structure_optimization_params
        )

        self.model.to(self.device)
        params = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optimizer(params, **optimizer_params)
        self.scheduler = StepLR(self.optimizer, **scheduler_params)

    def get_optimizable_module(self) -> torch.nn.Module:
        """Return the module for optimization applying."""
        return self.model.backbone

    def train_loop(self):
        """Training method of the model.

        Returns:
            Dictionary {metric_name: value}.
        """
        self.model.train()
        batches = tqdm(self.train_dl)
        train_scores = {'loss': 0}
        for key in self.structure_optimization.losses:
            train_scores[key] = 0
        i = 0
        for images, targets in batches:
            i += 1
            images = [image.to(self.device) for image in images]
            targets = [{k: v.to(self.device) for k, v in t.items()} for t in targets]
            loss_dict = self.model(images, targets)
            losses = sum(loss_dict.values())
            train_scores['loss'] += losses.item()

            for key, loss_fn in self.structure_optimization.losses.items():
                opt_loss = loss_fn(self.get_optimizable_module())
                losses += opt_loss
                train_scores[key] += opt_loss.item()
            self.optimizer.zero_grad()
            losses.backward()
            self.optimizer.step()
            batches.set_postfix({'loss': train_scores['loss'] / i})

        for key in train_scores:
            train_scores[key] /= len(self.train_dl)
        return train_scores

    def val_loop(self):
        """Validation method of the model.

        Returns:
            Dictionary {metric_name: value}.
        """
        metric = MeanAveragePrecision()
        tk = tqdm(self.val_dl)
        for images, targets in tk:
            images = list(image.to(self.device) for image in images)
            preds = self.forward(images)
            metric.update(preds, targets)
        return metric.compute()

    def _predict_on_image(self, image: torch.Tensor, proba: bool) -> Dict:
        """Returns prediction for image."""
        pred = self.forward([image])[0]
        if not proba:
            not_thresh = pred['scores'] > 0.5
            pred['boxes'] = pred['boxes'][not_thresh]
            pred['labels'] = pred['labels'][not_thresh]
            pred.pop('scores')
        for key in pred:
            pred[key] = pred[key].tolist()
        return pred

    def forward(self, images):
        """Predict model outputs from images.

        Args:
            images: List of image tensors.
        """
        self.model.eval()
        with torch.no_grad():
            images = list(image.to(self.device) for image in images)
            outputs = self.model(images)
            outputs = [{k: v.to('cpu').detach() for k, v in t.items()} for t in outputs]
        return outputs



import numpy as np
import pandas as pd
from sklearn.decomposition import PCA


class VarianceSelector:
    """
    Class that accepts a dictionary as input, the keys of which are the names of models and the values are arrays
    of data in the np.array format.The class implements an algorithm to determine the "best" set of features and the
    best model in the dictionary.
    """

    def __init__(self, models):
        """
        Initialize the class with the models dictionary.
        """
        self.models = models
        self.principal_components = {}
        self.model_scores = {}

    def get_best_model(self, **model_hyperparams):
        """
        Method to determine the "best" set of features and the best model in the dictionary.
        As an estimation algorithm, use the Principal Component analysis method and the proportion of the explained variance.
        If there are several best models, then a model with a smaller number of principal components and a
        larger value of the explained variance is chosen.
        """
        best_model = None
        best_score = 0
        for model_name, model_data in self.models.items():
            pca = PCA()
            pca.fit(model_data)
            filtred_score = [x for x in pca.explained_variance_ratio_ if x > 0.05]
            score = sum(filtred_score)
            self.principal_components.update({model_name: pca.components_[:, :len(filtred_score)]})
            self.model_scores.update({model_name: (score, len(filtred_score))})
            if score > best_score:
                best_score = score
                best_model = model_name
        return best_model

    def transform(self,
                  model_data,
                  principal_components):
        if type(principal_components) == str:
            principal_components = self.principal_components[principal_components]
        projected = np.dot(model_data, principal_components)
        return projected

    def select_discriminative_features(self,
                                       model_data,
                                       projected_data,
                                       corellation_level: float = 0.8):
        discriminative_feature = {}
        for PCT in range(projected_data.shape[1]):
            correlation_df = pd.DataFrame.corrwith(model_data, pd.Series(projected_data[:, PCT]), axis=0, drop=False)
            discriminative_feature_list = [k for k, x in zip(correlation_df.index.values, correlation_df.values) if
                                           abs(x) > corellation_level]
            discriminative_feature.update({f'{PCT + 1} principal components': discriminative_feature_list})
        return discriminative_feature



from fedot_ind.core.operation.decomposition.matrix_decomposition.DMDecomposition import *


class piDMD:
    """
    Computes a dynamic mode decomposition when the solution matrix is
    constrained to lie in a matrix manifold.
    The options available for the "method"
    - "exact", "exactSVDS"
    - "orthogonal"
    - "uppertriangular, "lowertriangular"
    - "diagonal", "diagonalpinv", "diagonaltls", "symtridiagonal"
    - "circulant", "circulantTLS", "circulantunitary", "circulantsymmetric",
    "circulantskewsymmetric"
    - "hankel", "toeplitz"
    - "symmetric", "skewsymmetric"
    """

    def __init__(self, method):
        self.method_dict = {"exact": exact_dmd_decompose,
                            "orthogonal": orthogonal_dmd_decompose,
                            'hankel': hankel_decompose,
                            'symmetric': symmetric_decompose
                            }
        self.method = method

    def fit(self, train_features, train_target):
        nx, nt = train_features.shape
        rank = min(nx, nt)
        self.fitted_linear_operator, eigenvals, eigenvectors = self.method_dict[self.method](train_features, train_target, rank)
        return self.fitted_linear_operator, eigenvals, eigenvectors

    def predict(self, test_features):
        return self.fitted_linear_operator(test_features)
        # elif method == 'uppertriangular':
        #     R, Q = rq(X)
        #     Ut = np.triu(Y @ Q.T)
        #     A = Ut / R
        # elif method == 'lowertriangular':
        #     A = np.rot90(piDMD(np.flipud(X), np.flipud(Y), 'uppertriangular'), 2)
        # elif method.startswith('diagonal'):
        #     if len(args) > 0:
        #         d = args[0]
        #         if len(d) == 1:
        #             d = d * np.ones((nx, 2))
        #         elif len(d) == nx:
        #             d = np.repeat(d[:, np.newaxis], 2, axis=1)
        #         elif any(d.shape != (nx, 2)):
        #             raise ValueError('Diagonal number is not in an allowable format.')
        #     else:
        #         d = np.ones((nx, 2))
        #     Icell = [None] * nx
        #     Jcell = [None] * nx
        #     Rcell = [None] * nx
        #     for j in range(nx):
        #         l1 = max(j - (d[j, 0] - 1), 0)
        #         l2 = min(j + (d[j, 1] - 1), nx - 1)
        #         C = X[l1:l2 + 1, :]
        #         b = Y[j, :]
        #         if method == 'diagonal':
        #             sol = b / C
        #         elif method == 'diagonalpinv':
        #             sol = b @ np.linalg.pinv(C)
        #         elif method == 'diagonaltls':
        #             sol = tls(C.T, b.T).T
        #         Icell[j] = j * np.ones(1 + l2 - l1)
        #         Jcell[j] = np.arange(l1, l2 + 1)
        #         Rcell[j] = sol
        #     Imat = np.concatenate(Icell)
        #     Jmat = np.concatenate(Jcell)
        #     Rmat = np.concatenate(Rcell)
        #     Asparse = scipy.sparse.csr_matrix((Rmat, (Imat, Jmat)), shape=(nx, nx))
        #     A = lambda v: Asparse @ v
        #     if nargout == 2:
        #         eVals = eigs(Asparse, nx)
        #         varargout[0] = eVals
        #     else:
        #         eVecs, eVals = eigs(Asparse, nx)
        #         varargout[0] = np.diag(eVals)
        #         varargout[1] = eVecs
        # elif method == 'symmetric' or method == 'skewsymmetric':
        #     Ux, S, V = svd(X, 0)
        #     C = Ux.T @ Y @ V
        #     C1 = C
        #     if len(args) > 0:
        #         r = args[0]
        #     else:
        #         r = np.linalg.matrix_rank(X)



from fedot.api.main import Fedot


class OneStageBooster:
    def __init__(self, true_target, base_predict, features):
        self.true_target = true_target
        self.base_predict = base_predict
        self.features = features
        self.model = None
        self.model_hyperparams = dict(problem='regression', seed=42,
                                      timeout=5, max_depth=10,
                                      max_arity=4, cv_folds=2,
                                      logging_level=20, n_jobs=2)

    def fit(self):
        self.model = Fedot(**self.model_hyperparams)
        diff_target = self.true_target - self.base_predict
        self.model.fit(self.features, diff_target)

    def predict(self, features):
        return self.model.predict(features)

    def score(self, predict, target):
        ...




import numpy as np
from hyperopt import hp

industrial_search_space = {
    'data_driven_basis':
        {'n_components': (hp.uniformint, [2, 10]),
         'window_size': (hp.uniformint, [10, 50])},
    'wavelet_basis':
        {'n_components': (hp.uniformint, [2, 10]),
         'wavelet': (hp.choice, [['mexh', 'shan', 'morl', 'cmor', 'fbsp', 'db5', 'sym5']])},
    'fourier_basis':
        {'spectrum': (hp.choice, [['smoothed']]),
         'threshold':(hp.uniformint, [10000, 50000])},
    'quantile_extractor':
        {'window_size': (hp.uniformint, [1, 50]),
         'win_mode': (hp.choice, [[True, False]])},
    'recurrence_extractor':
        {'window_size': (hp.uniformint, [1, 50]),
         'win_mode': (hp.choice, [[True, False]]),
         'min_signal_ratio': (hp.uniform, [0, 0.5]),
         'max_signal_ratio': (hp.uniform, [0.5, 1]),
         'rec_metric': (hp.choice, [['chebyshev', 'cosine', 'euclidean' 'mahalanobis']])},
    'signal_extractor':
        {'wavelet': (hp.choice, [['mexh', 'shan', 'morl', 'cmor', 'fbsp', 'db5', 'sym5']])}
}


def get_industrial_search_space(self):
    parameters_per_operation = {
        'kmeans': {
            'n_clusters': (hp.uniformint, [2, 7])
        },
        'adareg': {

            'learning_rate': (hp.loguniform, [np.log(1e-3), np.log(1)]),
            'loss': (hp.choice, [["linear", "square", "exponential"]])
        },
        'gbr': {

            'loss': (hp.choice, [["ls", "lad", "huber", "quantile"]]),
            'learning_rate': (hp.loguniform, [np.log(1e-3), np.log(1)]),
            'max_depth': (hp.uniformint, [1, 11]),
            'min_samples_split': (hp.uniformint, [2, 21]),
            'min_samples_leaf': (hp.uniformint, [1, 21]),
            'subsample': (hp.uniform, [0.05, 1.0]),
            'max_features': (hp.uniform, [0.05, 1.0]),
            'alpha': (hp.uniform, [0.75, 0.99])
        },
        'logit': {
            'C': (hp.uniform, [1e-2, 10.0])
        },
        'rf': {
                'criterion': (hp.choice, [["gini", "entropy"]]),
                'max_features': (hp.uniform, [0.05, 1.0]),
                'min_samples_split': (hp.uniformint, [2, 10]),
                'min_samples_leaf': (hp.uniformint, [1, 15]),
                'bootstrap': (hp.choice, [[True, False]])
            },
        'lasso': {
            'alpha': (hp.uniform, [0.01, 10.0])
        },
        'ridge': {
            'alpha': (hp.uniform, [0.01, 10.0])
        },
        'rfr': {

            'max_features': (hp.uniform, [0.05, 1.0]),
            'min_samples_split': (hp.uniformint, [2, 21]),
            'min_samples_leaf': (hp.uniformint, [1, 21]),
            'bootstrap': (hp.choice, [[True, False]])
        },
        'xgbreg': {

            'max_depth': (hp.uniformint, [1, 11]),
            'learning_rate': (hp.loguniform, [np.log(1e-3), np.log(1)]),
            'subsample': (hp.uniform, [0.05, 1.0]),
            'min_child_weight': (hp.uniformint, [1, 21]),
            'objective': (hp.choice, [['reg:squarederror']])
        },
        'xgboost': {

            'max_depth': (hp.uniformint, [1, 7]),
            'learning_rate': (hp.loguniform, [np.log(1e-3), np.log(1)]),
            'subsample': (hp.uniform, [0.05, 0.99]),
            'min_child_weight': (hp.uniform, [1, 21])
        },
        'svr': {
            'loss': (hp.choice, [["epsilon_insensitive", "squared_epsilon_insensitive"]]),
            'tol': (hp.loguniform, [np.log(1e-5), np.log(1e-1)]),
            'C': (hp.uniform, [1e-4, 25.0]),
            'epsilon': (hp.uniform, [1e-4, 1.0])
        },
        'dtreg': {
            'max_depth': (hp.uniformint, [1, 11]),
            'min_samples_split': (hp.uniformint, [2, 21]),
            'min_samples_leaf': (hp.uniformint, [1, 21])
        },
        'treg': {

            'max_features': (hp.uniform, [0.05, 1.0]),
            'min_samples_split': (hp.uniformint, [2, 21]),
            'min_samples_leaf': (hp.uniformint, [1, 21]),
            'bootstrap': (hp.choice, [[True, False]])
        },
        'dt': {
            'max_depth': (hp.uniformint, [1, 11]),
            'min_samples_split': (hp.uniformint, [2, 21]),
            'min_samples_leaf': (hp.uniformint, [1, 21])
        },
        'knnreg': {
            'n_neighbors': (hp.uniformint, [1, 50]),
            'weights': (hp.choice, [["uniform", "distance"]]),
            'p': (hp.choice, [[1, 2]])
        },
        'knn': {
            'n_neighbors': (hp.uniformint, [1, 50]),
            'weights': (hp.choice, [["uniform", "distance"]]),
            'p': (hp.choice, [[1, 2]])
        },
        'arima': {
            'p': (hp.uniformint, [1, 7]),
            'd': (hp.uniformint, [0, 2]),
            'q': (hp.uniformint, [1, 5])
        },
        'stl_arima': {
            'p': (hp.uniformint, [1, 7]),
            'd': (hp.uniformint, [0, 2]),
            'q': (hp.uniformint, [1, 5]),
            'period': (hp.uniformint, [1, 365])
        },
        'ar': {
            'lag_1': (hp.uniform, [2, 200]),
            'lag_2': (hp.uniform, [2, 800])
        },
        'ets': {
            'error': (hp.choice, [['add', 'mul']]),
            'trend': (hp.choice, [[None, 'add', 'mul']]),
            'seasonal': (hp.choice, [[None, 'add', 'mul']]),
            'damped_trend': (hp.choice, [[True, False]]),
            'seasonal_periods': (hp.uniform, [1, 100])
        },
        'glm': {'nested_space': (hp.choice, [[
            {
                'family': 'gaussian',
                'link': hp.choice('link_gaussian', ['identity',
                                                    'inverse_power',
                                                    'log'])
            },
            {
                'family': 'gamma',
                'link': hp.choice('link_gamma', ['identity',
                                                 'inverse_power',
                                                 'log'])
            },
            {
                'family': 'inverse_gaussian',
                'link': hp.choice('link_inv_gaussian', ['identity',
                                                        'inverse_power'])
            }

        ]])},
        'clstm': {
            'window_size': (hp.uniform, [1, 200]),
            'hidden_size': (hp.uniform, [20, 200]),
            'teacher_forcing': (hp.uniform, [0, 1]),
            'learning_rate': (hp.uniform, [0.0005, 0.005]),
            'cnn1_kernel_size': (hp.uniformint, [3, 8]),
            'cnn1_output_size': (hp.choice, [[8, 16, 32, 64]]),
            'cnn2_kernel_size': (hp.uniformint, [3, 8]),
            'cnn2_output_size': (hp.choice, [[8, 16, 32, 64]]),
            'batch_size': (hp.choice, [[64, 128]]),
            'num_epochs': (hp.choice, [[10, 20, 50, 100]]),
            'optimizer': (hp.choice, [['adam', 'sgd']]),
            'loss': (hp.choice, [['mae', 'mse']])
        },
        'pca': {
            'n_components': (hp.uniform, [0.1, 0.99]),
            'svd_solver': (hp.choice, [['full']])
        },
        'kernel_pca': {
            'n_components': (hp.uniformint, [1, 20]),
            'kernel': (hp.choice, [['linear', 'poly', 'rbf', 'sigmoid', 'cosine', 'precomputed']])
        },
        'fast_ica': {
            'n_components': (hp.uniformint, [1, 20]),
            'fun': (hp.choice, [['logcosh', 'exp', 'cube']])
        },
        'ransac_lin_reg': {
            'min_samples': (hp.uniform, [0.1, 0.9]),
            'residual_threshold': (hp.loguniform, [np.log(0.1), np.log(1000)]),
            'max_trials': (hp.uniform, [50, 500]),
            'max_skips': (hp.uniform, [50, 500000])
        },
        'ransac_non_lin_reg': {
            'min_samples': (hp.uniform, [0.1, 0.9]),
            'residual_threshold': (hp.loguniform, [np.log(0.1), np.log(1000)]),
            'max_trials': (hp.uniform, [50, 500]),
            'max_skips': (hp.uniform, [50, 500000])
        },
        'isolation_forest_reg': {
            'max_samples': (hp.uniform, [0.05, 0.99]),
            'max_features': (hp.uniform, [0.05, 0.99]),
            'bootstrap': (hp.choice, [[True, False]])
        },
        'isolation_forest_class': {
            'max_samples': (hp.uniform, [0.05, 0.99]),
            'max_features': (hp.uniform, [0.05, 0.99]),
            'bootstrap': (hp.choice, [[True, False]])
        },
        'rfe_lin_reg': {
            'n_features_to_select': (hp.uniform, [0.5, 0.9]),
            'step': (hp.uniform, [0.1, 0.2])
        },
        'rfe_non_lin_reg': {
            'n_features_to_select': (hp.uniform, [0.5, 0.9]),
            'step': (hp.uniform, [0.1, 0.2])
        },
        'poly_features': {
            'degree': (hp.uniformint, [2, 5]),
            'interaction_only': (hp.choice, [[True, False]])
        },
        'polyfit': {
            'degree': (hp.uniformint, [1, 6])
        },
        'lagged': {
            'window_size': (hp.uniformint, [5, 500])
        },
        'sparse_lagged': {
            'window_size': (hp.uniformint, [5, 500]),
            'n_components': (hp.uniform, [0, 0.5]),
            'use_svd': (hp.choice, [[True, False]])
        },
        'smoothing': {
            'window_size': (hp.uniformint, [2, 20])
        },
        'gaussian_filter': {
            'sigma': (hp.uniform, [1, 5])
        },
        'diff_filter': {
            'poly_degree': (hp.uniformint, [1, 5]),
            'order': (hp.uniform, [1, 3]),
            'window_size': (hp.uniform, [3, 20])
        },
        'cut': {
            'cut_part': (hp.uniform, [0, 0.9])
        },
        'lgbm': {
            'class_weight': (hp.choice, [[None, 'balanced']]),
            'num_leaves': (hp.uniformint, [2, 256]),
            'learning_rate': (hp.loguniform, [np.log(0.01), np.log(0.2)]),
            'colsample_bytree': (hp.uniform, [0.4, 1]),
            'subsample': (hp.uniform, [0.4, 1]),
            'reg_alpha': (hp.loguniform, [np.log(1e-8), np.log(10)]),
            'reg_lambda': (hp.loguniform, [np.log(1e-8), np.log(10)])
        },
        'lgbmreg': {
            'num_leaves': (hp.uniformint, [2, 256]),
            'learning_rate': (hp.loguniform, [np.log(0.01), np.log(0.2)]),
            'colsample_bytree': (hp.uniform, [0.4, 1]),
            'subsample': (hp.uniform, [0.4, 1]),
            'reg_alpha': (hp.loguniform, [np.log(1e-8), np.log(10)]),
            'reg_lambda': (hp.loguniform, [np.log(1e-8), np.log(10)])
        },
        'catboost': {
            'max_depth': (hp.uniformint, [1, 11]),
            'learning_rate': (hp.loguniform, [np.log(0.01), np.log(0.2)]),
            'min_data_in_leaf': (hp.qloguniform, [0, 6, 1]),
            'border_count': (hp.uniformint, [2, 255]),
            'l2_leaf_reg': (hp.loguniform, [np.log(1e-8), np.log(10)]),
            'loss_function': (hp.choice, [['Logloss', 'CrossEntropy']])
        },
        'catboostreg': {
            'max_depth': (hp.uniformint, [1, 11]),
            'learning_rate': (hp.loguniform, [np.log(0.01), np.log(0.2)]),
            'min_data_in_leaf': (hp.qloguniform, [0, 6, 1]),
            'border_count': (hp.uniformint, [2, 255]),
            'l2_leaf_reg': (hp.loguniform, [np.log(1e-8), np.log(10)])
        },
        'resample': {
            'balance': (hp.choice, [['expand_minority', 'reduce_majority']]),
            'replace': (hp.choice, [[True, False]]),
            'balance_ratio': (hp.uniform, [0.3, 1])
        },
        'lda': {
            'solver': (hp.choice, [['svd', 'lsqr', 'eigen']]),
            'shrinkage': (hp.uniform, [0.1, 0.9])
        },
        'ts_naive_average': {'part_for_averaging': (hp.uniform, [0.1, 1])},
        'locf': {'part_for_repeat': (hp.uniform, [0.01, 0.5])},
        'word2vec_pretrained': {
            'model_name': (hp.choice, [['glove-twitter-25', 'glove-twitter-50',
                                        'glove-wiki-gigaword-100', 'word2vec-ruscorpora-300']])
        },
        'tfidf': {
            'ngram_range': (hp.choice, [[(1, 1), (1, 2), (1, 3)]]),
            'min_df': (hp.uniform, [0.0001, 0.1]),
            'max_df': (hp.uniform, [0.9, 0.99])
        }, **industrial_search_space
    }

    if self.custom_search_space is not None:
        for operation in self.custom_search_space.keys():
            if self.replace_default_search_space:
                parameters_per_operation[operation] = self.custom_search_space[operation]
            else:
                for key, value in self.custom_search_space[operation].items():
                    parameters_per_operation[operation][key] = value

    return parameters_per_operation



import os
from pathlib import Path

PROJECT_PATH = str(Path(__file__).parent.parent.parent.parent)


def default_path_to_save_results() -> str:
    path = PROJECT_PATH
    save_path = os.path.join(path, 'results_of_experiments')
    return save_path



import pandas as pd
from pymonad.list import ListMonad
from sklearn.preprocessing import MinMaxScaler

from fedot_ind.api.utils.hp_generator_collection import GeneratorParams
from fedot_ind.core.architecture.postprocessing.Analyzer import PerformanceAnalyzer
from fedot_ind.core.architecture.settings.pipeline_factory import BasisTransformations, FeatureGenerator, MlModel


class AbstractPipelines:
    def __init__(self, train_data, test_data):
        self.train_features = train_data[0]
        self.train_target = train_data[1]
        self.test_features = test_data[0]
        self.test_target = test_data[1]
        self.basis = None

        self.basis_dict = {i.name: i.value for i in BasisTransformations}
        self.model_dict = {i.name: i.value for i in MlModel}
        self.feature_generator_dict = {i.name: i.value for i in FeatureGenerator}

        self.generators_with_matrix_input = ['topological',
                                             'wavelet',
                                             'reccurence',
                                             'statistical']

    def _evaluate(self, classificator, train_features, test_features):
        fitted_model = classificator.fit(train_features=train_features,
                                         train_target=self.train_target)
        predicted_probs_labels = (classificator.predict(test_features=test_features),
                                  classificator.predict_proba(test_features=test_features))
        metrics = PerformanceAnalyzer().calculate_metrics(target=self.test_target,
                                                          predicted_labels=predicted_probs_labels[0],
                                                          predicted_probs=predicted_probs_labels[1])
        return fitted_model, metrics

    def get_feature_generator(self, **kwargs):
        pass

    def _get_feature_matrix(self, list_of_features, mode: str = 'Multi', **kwargs):
        if mode == '1D':
            feature_matrix = pd.concat(list_of_features, axis=0)
            if feature_matrix.shape[0] != len(list_of_features):
                feature_matrix = pd.concat(list_of_features, axis=1)
        elif mode == 'MultiEnsemble':
            feature_matrix = []
            for i in range(len(list_of_features[0])):
                _ = []
                for feature_set in list_of_features:
                    _.append(feature_set[i])
                feature_matrix.append(pd.concat(_, axis=0))
        elif mode == 'list_of_ts':
            feature_matrix = []
            for ts in list_of_features:
                list_of_windows = []
                for step in range(0, ts.shape[1], kwargs['window_length']):
                    list_of_windows.append(ts[:, step:step + kwargs['window_length']])
                feature_matrix.append(list_of_windows)
        else:
            feature_matrix = pd.concat([pd.concat(feature_set, axis=1) for feature_set in list_of_features], axis=0)
        return feature_matrix

    def _init_pipeline_nodes(self, model_type: str = 'tsc', **kwargs):
        if 'feature_generator_type' not in kwargs.keys():
            generator = self.feature_generator_dict['statistical']
        else:
            generator = self.feature_generator_dict[kwargs['feature_generator_type']]
        try:
            feature_extractor = generator(params=kwargs['feature_hyperparams'])

        except AttributeError:
            params = GeneratorParams[kwargs['feature_generator_type']].value
            feature_extractor = generator(params)
        classificator = self.model_dict[model_type](model_hyperparams=kwargs['model_hyperparams'],
                                                    generator_name=kwargs['feature_generator_type'],
                                                    generator_runner=feature_extractor)
    # TODO:
        lambda_func_dict = {'create_list_of_ts': lambda x: ListMonad(*x.values.tolist()),
                            'scale': lambda time_series: pd.DataFrame(MinMaxScaler().fit_transform(
                                time_series.to_numpy())),
                            'transpose_matrix': lambda time_series: time_series.T,
                            'reduce_basis': lambda x: x[:, 0] if x.shape[1] == 1 else x[:, kwargs['component']],
                            'extract_features': lambda x: feature_extractor.get_features(x),
                            'fit_model': lambda x: classificator.fit(train_features=x, train_target=self.train_target),
                            'predict': lambda x: ListMonad({'predicted_labels': classificator.predict(test_features=x),
                                                            'predicted_probs': classificator.predict_proba(
                                                                test_features=x)})
                            }

        return feature_extractor, classificator, lambda_func_dict



from enum import Enum


class GeneratorParams(Enum):
    quantile = {'window_mode': True,
                'use_cache': False,
                'window_size': 20}

    statistical = {'window_mode': False,
                   'use_cache': False}

    recurrence = {'use_cache': False,
                  'threshold_baseline': [1, 5, 10, 15, 20, 25, 30],
                  'min_signal_ratio': 0.65,
                  'max_signal_ratio': 0.75,
                  'rec_metric': 'euclidean'}

    topological = {'use_cache': False,
                   'max_te_dimension': 5,
                   'max_te_time_delay': 2,
                   'stride': 1}

    spectral = {'window_sizes': [],
                'window_mode': True,
                'use_cache': False,
                'correlation_level': 0.8,
                'combine_eigenvectors': False}

    ensemble = {'list_of_generators': {},
                'use_cache': False}

    window_spectral = {'window_mode': True,
                       'window_sizes': [],
                       'use_cache': False}

    wavelet = {'use_cache': False,
               'wavelet': 'mexh'}



import logging
from typing import Union

import numpy as np
import pandas as pd


class ParameterCheck:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def check_window_sizes(self, config_dict: dict,
                           dataset_name: str,
                           train_data: Union[pd.DataFrame, np.ndarray]):

        for generator in config_dict['feature_generator']:
            if generator in ['spectral', 'window_spectral']:
                self.logger.info(f'Check window sizes for {generator} generator and {dataset_name} dataset')
                if dataset_name not in config_dict['feature_generator_params'][generator]['window_sizes'].keys():

                    ts_length = train_data[0].shape[1]
                    window_sizes_list = list(map(lambda x: round(ts_length / x), [10, 5, 3]))
                    config_dict['feature_generator_params'][generator]['window_sizes'][dataset_name] = window_sizes_list
                    self.logger.info(f'Window sizes for {dataset_name} are not specified. '
                                     f'Auto-selected: {window_sizes_list}')
                else:
                    self.logger.info(f'Window sizes for dataset {dataset_name} are predefined')
            else:
                continue

        return config_dict

    def check_metric_type(self, target):
        n_classes = np.unique(target).shape[0]
        if n_classes > 2:
            self.logger.info('Metric for evaluation - F1')
            return 'f1'
        else:
            self.logger.info('Metric for evaluation - ROC-AUC')
            return 'roc_auc'

    def check_baseline_type(self, config_dict: None, model_params: dict):
        if config_dict is not None:
            baseline_type = config_dict['baseline']
        elif 'baseline_type' in model_params.keys():
            baseline_type = model_params['baseline']
        else:
            baseline_type = None
        return baseline_type

    def get_ds_and_generator_combinations(self, datasets, generators):
        combinations = []
        for dataset in datasets:
            for generator in generators:
                combinations.append((dataset, generator))
        return combinations


class DataCheck:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def _replace_inf_with_nans(input_data: np.ndarray):
        values_to_replace = [np.inf, -np.inf]
        features_with_replaced_inf = np.where(np.isin(input_data,
                                                      values_to_replace),
                                              np.nan,
                                              input_data)
        input_data = features_with_replaced_inf
        return input_data

    def _check_for_nan(self, input_data: np.ndarray) -> np.ndarray:
        """Method responsible for checking if there are any NaN values in the time series dataframe
        and replacing them with 0

        Args:
            input_data: time series dataframe with NaN values

        Returns:
            input_data: time series dataframe without NaN values

        """
        if np.any(np.isnan(input_data)):
            input_data = np.nan_to_num(input_data, nan=0)
        return input_data

    def check_data(self, input_data: pd.DataFrame, return_df: bool = False) -> Union[pd.DataFrame, np.ndarray]:
        if type(input_data) != pd.DataFrame:
            return input_data
        else:
            filled_data = input_data.apply(lambda x: self._replace_inf_with_nans(x))
            filled_data = filled_data.apply(lambda x: self._check_for_nan(x))
            if filled_data.shape[0] == input_data.shape[0] and filled_data.shape[1] == input_data.shape[1]:
                input_data = filled_data
            else:
                self.logger.error('Encontered error during extracted features checking')
                raise ValueError('Data contains NaN values')

            if return_df:
                return input_data
            else:
                return input_data.values



from typing import Optional
import numpy as np
import pywt
from fedot.core.operations.operation_parameters import OperationParameters
from pymonad.either import Either
from pymonad.list import ListMonad
from fedot_ind.core.operation.transformation.basis.abstract_basis import BasisDecompositionImplementation


class WaveletBasisImplementation(BasisDecompositionImplementation):
    """Wavelet basis
        Example:
            ts1 = np.random.rand(200)
            ts2 = np.random.rand(200)
            ts = [ts1, ts2]
            bss = WaveletBasisImplementation({'n_components': 2, 'wavelet': 'mexh'})
            basis_multi = bss._transform(ts)
            basis_1d = bss._transform(ts1)
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.n_components = params.get('n_components')
        self.wavelet = params.get('wavelet')
        self.basis = None
        self.discrete_wavelets = pywt.wavelist(kind='discrete')
        self.continuous_wavelets = pywt.wavelist(kind='continuous')
        self.scales = [2, 4, 10, 20]

    def _decompose_signal(self, input_data):
        if self.wavelet in self.discrete_wavelets:
            high_freq, low_freq = pywt.dwt(input_data, self.wavelet, 'smooth')
        else:
            high_freq, low_freq = pywt.cwt(data=input_data,
                                           scales=self.scales,
                                           wavelet=self.wavelet)
            low_freq = high_freq[-1, :]
            high_freq = np.delete(high_freq, (-1), axis=0)
            low_freq =low_freq[np.newaxis, :]
        return high_freq, low_freq

    def _decomposing_level(self):
        """The level of decomposition of the time series.

        Returns:
            The level of decomposition of the time series.
        """
        return pywt.dwt_max_level(len(self.time_series), self.wavelet)

    def _transform_one_sample(self, series: np.array):
        return self._get_basis(series)

    def _get_1d_basis(self, data):

        decompose = lambda signal: ListMonad(self._decompose_signal(signal))
        threshold = lambda Monoid: ListMonad([Monoid[0][
                                              :self.n_components],
                                              Monoid[1]])

        basis = Either.insert(data).then(decompose).then(threshold).value[0]
        basis = np.concatenate(basis)
        return basis

    def _get_multidim_basis(self, data):
        decompose = lambda multidim_signal: ListMonad(list(map(self._decompose_signal, multidim_signal)))
        select_level = lambda Monoid: [Monoid[0][
                                       :self.n_components, :],
                                       Monoid[1]]
        threshold = lambda decomposed_signal: list(map(select_level, decomposed_signal))

        basis = Either.insert(data).then(decompose).then(threshold).value
        return basis





import numpy as np
from scipy.linalg import solve_triangular
from sklearn.decomposition import PCA
from fedot_ind.core.operation.filtration.quantile_filtration import quantile_filter
# from core.operation.transformation.regularization.lp_reg import compute_penalty_matrix


class FunctionalPCA:
    """
    Principal component analysis.
    Class that implements functional principal component analysis for both
    basis and natural representations of the data.

    Parameters:
        model_hyperparams:
                        n_components: Number of principal components to keep from
                            functional principal component analysis. Defaults to 3.
                        regularization: Regularization object to be applied.
                        basis_of_function: .
    Attributes:
        components\_: this contains the principal components.
        explained_variance\_ : The amount of variance explained by
            each of the selected components.
        explained_variance_ratio\_ : this contains the percentage
            of variance explained by each principal component.
        singular_values\_: The singular values corresponding to each of the
            selected components.
        mean\_: mean of the train data.
    Examples:

    time_series = np.array([1,2,3,4,5,6])
    data_range = len(time_series)
    basis = ChebyshevBasis(data_range=data_range, n_components=4).decompose(time_series)
    FPCA = FunctionalPCA(2)
    FPCA = FPCA.fit(basis)

    """

    def __init__(
            self, model_hyperparams: dict = None
    ) -> None:
        self.mean_ = 0
        self.n_components = 2
        self.regularization = None
        self.basis_function = None

        if model_hyperparams is not None:
            self.n_components = model_hyperparams['n_components']
            self.regularization = model_hyperparams['regularization']
            # self._weights = model_hyperparams['weights']
            self.basis_function = model_hyperparams['basis_function']

    def _delete_mean(self, X):
        if type(X) is not np.ndarray:
            X = X.values
        mean_centred = np.allclose(X, X - np.mean(X, axis=0))
        if not mean_centred:
            self.mean_ = np.mean(X, axis=0)
            X = X - self.mean_
        return X

    def _fit_basis(
            self,
            X: np.ndarray
    ):
        """
        Compute the first n_components principal components and saves them.
        Args:
            X: The functional data object to be analysed.
            y: Ignored.
        Returns:
            self
        References:
            .. [RS05-8-4-2] Ramsay, J., Silverman, B. W. (2005). Basis function
                expansion of the functions. In *Functional Data Analysis*
                (pp. 161-164). Springer.
        """
        X = self._delete_mean(X)
        if self.basis_function is not None:
            # Compute Gram matrix of basis function
            G = self.basis_function.T.dot(self.basis_function)
            # The matrix that are in charge of changing the computed principal
            # components to target matrix is essentially the inner product
            # of both basis.
            J = np.dot(X, self.basis_function)
        else:
            # If no other basis is specified we use the same basis as the
            components_basis = X.copy()
            # G = pairwise_distance(X.T)
            G = X.T.dot(X)
            J = G

        self._X_basis = X
        self._j_matrix = J

        # # Apply regularization
        # if self.regularization is not None:
        #     regularization_matrix = compute_penalty_matrix(
        #         basis_iterable=(components_basis,),
        #         regularization_parameter=1,
        #         regularization=self.regularization,
        #     )
        #
        #     G = G + regularization_matrix

        # Diagonalisation of Gram Matrix. G = L*L^T
        l_matrix = np.linalg.cholesky(G)

        # we need L^{-1} for a multiplication, there are two possible ways:
        # using solve to get the multiplication result directly or just invert
        # the matrix. We choose solve because it is faster and more stable.
        # The following matrix is needed: L^{-1}*J^T
        l_inv_j_t = solve_triangular(
            l_matrix,
            np.transpose(J),
            lower=True,
        )

        # the final matrix, C(L-1Jt)t for svd or (L-1Jt)-1CtC(L-1Jt)t for PCA

        final_matrix = X @ np.transpose(l_inv_j_t)

        # initialize the pca module provided by scikit-learn
        pca = PCA(n_components=self.n_components)
        pca.fit(final_matrix)

        # we choose solve to obtain the component coefficients for the
        # same reason: it is faster and more efficient
        component_coefficients = solve_triangular(
            np.transpose(l_matrix),
            np.transpose(pca.components_),
            lower=False,
        )

        self.pca = pca
        self.component_coefficients = component_coefficients
        self.components_ = component_coefficients.T @ X.T

        return self

    def _transform_basis(
            self,
            X: np.array
    ):
        """Compute the n_components first principal components score.
        Args:
            X: The functional data object to be analysed.
        Returns:
            Principal component scores.
        """

        # Compute inner product of our data with the components
        return X @ self._j_matrix @ self.component_coefficients

    def fit(
            self,
            X: np.array
    ):
        """
        Compute the n_components first principal components and saves them.
        Args:
            X: The functional data object to be analysed.
        Returns:
            self
        """

        return self._fit_basis(X)

    def transform(
            self,
            X: np.array
    ):
        """
        Compute the ``n_components`` first principal components scores.
        Args:
            X: The functional data object to be analysed.
        Returns:
            Principal component scores.
        """
        X = self._delete_mean(X)
        return self._transform_basis(X)

    def predict(self, test_features, threshold: float = 0.99):

        if type(test_features) == list:
            list_of_projection, list_of_outliers = [], []
            for window_slice in test_features:
                current_projection, current_outlier = self._predict(window_slice.T, threshold)
                list_of_projection.append(current_projection)
                list_of_outliers.append(current_outlier)
            return list_of_projection, list_of_outliers
        else:
            return self._predict(test_features.T, threshold)

    def _predict(self, test_features, threshold: float = 0.90):
        projection = self.transform(test_features)
        recover = self.inverse_transform(projection)
        outlier_idx = quantile_filter(input_data=test_features, predicted_data=recover)
        return recover, outlier_idx

    def fit_transform(
            self,
            X: np.array,
    ):
        """
        Compute the n_components first principal components and their scores.
        Args:
            X: The functional data object to be analysed.
        Returns:
            Principal component scores.
        """
        return self.fit(X).transform(X)

    def inverse_transform(
            self,
            pc_scores,
    ):
        """
        Compute the recovery from the fitted principal components scores.
        In other words,
        it maps ``pc_scores``, from the fitted functional PCs' space,
        back to the input functional space.
        ``pc_scores`` might be an array returned by ``transform`` method.
        Args:
            pc_scores: ndarray (n_samples, n_components).
        Returns:

        """

        reconstructed = pc_scores @ self.component_coefficients.T

        return reconstructed + self.mean_
